"""
This code was created for and sponsored by Cartiga (www.cartiga.com). 
Cartiga makes no representations or warranties in connection with the code 
and waives any and all liability in connection therewith. Your use of the 
code constitutes acceptance of these terms.

Copyright 2022 Cartiga LLC
"""

import surpyval as surv

from surpyval.utils import (
    _get_idx,
    validate_cr_inputs,
    # validate_cr_df_inputs,
    validate_event,
    validate_cif_event
)
import textwrap

from surpyval.nonparametric.nelson_aalen import nelson_aalen as na
from surpyval.nonparametric.kaplan_meier import kaplan_meier as km

class CompetingRisks():
    def __repr__(self):
        out = """\
        Competing Risk model with events:
        {events}
        """.format(events=list(self.event_idx_map.keys()))
        return textwrap.dedent(out)


    def _f(self, f, x, event):
        validate_event(self.event_idx_map, event)
        idx, rev = _get_idx(self.x, x)

        if f == 'h':
            arr = self.h0_e
        elif f == 'H':
            arr = self.H0_e
        elif f == "IIF":
            arr = self.IIF
        elif f =="CIF":
            arr = self.CIF

        if event is None:
            return arr.sum(axis=0)[idx][rev]
        else:
            e = self.event_idx_map[event]
            return arr[e, idx][rev]

    def hf(self, x, event=None):
        return self._f('h', x, event)

    def Hf(self, x, event=None):
        return self._f('H', x, event)

    def sf(self, x, event=None):
        return np.exp(-self.Hf(x, event=event))

    def ff(self, x, event=None):
        """
        A lot of commentary about this being difficult to interpret.
        In engineering this is not the case, eliminating the failure
        will result in the ff being gone.
        """
        return 1 - np.exp(-self.Hf(x, event=event))

    def df(self, x, event=None):
        return self.hf(x, event=event) * self.sf(x, event=event)

    def iif(self, x, event):
        """
        Instantaneous Incidence Function
        """
        validate_cif_event(event)
        return self._f('IIF', x, event)

    def cif(self, x, event):
        """
        Cumulative Incidence Function
        """
        validate_cif_event(event)

        return self._f('CIF', x, event)

    @classmethod
    def fit_from_df(cls, df, x_col, e_col, c_col=None, 
                    n_col=None, method="Nelson-Aalen"):

        x, c, n, e = validate_cr_df_inputs(df, x_col, e_col, 
                                           c_col, n_col)
        model = cls.fit(x, e, c, n, method)
        model.df = df
        return model

    @classmethod
    def fit(cls, x, e, c=None, n=None, method="Nelson-Aalen"):
        """
        Need to check that causes is the same length
        TODO: FlemingHarrington baseline.
        """
        x, c, n, e = validate_cr_inputs(x, c, n, e, method)

        # Get unique event types
        unique_e = set(e)
        # Remove None type, which relates to censored obs
        # np.unique doesn't work since it can't handle None
        if None in unique_e:
            unique_e.remove(None)

        # Count number of unique event types.
        # Ordering is stable over repeats due to sort.
        n_event_types = len(unique_e)
        event_idx_map = {state : i for i, state in enumerate(sorted(unique_e))}

        # Get the x, r, d format agnostic of event.
        unique_x, r, d = surv.xcn_to_xrd(x, c, n)

        # empty count array of occurrence (e) of amount (d) at time (x)
        d_e = np.zeros((n_event_types, len(unique_x)))

        # Counter for each occurrence
        for i, x_i in enumerate(x):
            if c[i] == 1:
                continue
            j = event_idx_map[e[i]]
            d_e[j, np.where(unique_x == x_i)] += n[i]

        if method == "Nelson-Aalen":
            S = na(r, d)
        elif method == "Kaplan-Meier":
            S = km(r, d)

        # Useful object to return to user
        model = cls()
        model.n_event_types = n_event_types
        model.event_idx_map = event_idx_map

        # Store relevant data to object
        model.x = unique_x
        model.d = d
        model.r = r
        model.h0 = d / r
        model.H0 = model.h0.cumsum()
        model.S = S
        model.d_e = d_e
        model.h0_e = d_e / r
        model.H0_e = model.h0_e.cumsum(axis=1)
        model.IIF = model.S * model.h0_e
        model.CIF = model.IIF.cumsum(axis=1)
        return model


from surpyval import np
from scipy.stats import uniform
from autograd import jacobian
from numpy import euler_gamma
from scipy.special import gamma as gamma_func
from scipy.special import ndtri as z
from scipy import integrate
from scipy.optimize import minimize

from surpyval import parametric as para
from surpyval import nonparametric as nonp
from surpyval.parametric.parametric_fitter import ParametricFitter

from .fitters.mpp import mpp

class ExpoWeibull_(ParametricFitter):
    def __init__(self, name):
        self.name = name
        self.k = 3
        self.bounds = ((0, None), (0, None), (0, None),)
        self.support = (0, np.inf)
        self.plot_x_scale = 'log'
        self.y_ticks = [0.0001, 0.0002, 0.0003, 0.001, 0.002, 
            0.003, 0.005, 0.01, 0.02, 0.03, 0.05, 
            0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 
            0.9, 0.95, 0.99, 0.999, 0.9999]
        self.param_names = ['alpha', 'beta', 'mu']
        self.param_map = {
            'alpha' : 0,
            'beta'  : 1,
            'mu'    : 2
        }

    def _parameter_initialiser(self, x, c=None, n=None, offset=False):
        log_x = np.log(x)
        log_x[np.isnan(log_x)] = 0
        gumb = para.Gumbel.fit(log_x, c, n, how='MLE')
        if not gumb.res.success:
            gumb = para.Gumbel.fit(log_x, c, n, how='MPP')
        mu, sigma = gumb.params
        alpha, beta = np.exp(mu), 1. / sigma
        if (np.isinf(alpha) | np.isnan(alpha)):
            alpha = np.median(x)
        if (np.isinf(beta) | np.isnan(beta)):
            beta = 1.
        if offset:
            gamma = np.min(x) - (np.max(x) - np.min(x))/10.
            return gamma, alpha, beta, 1.
        else:
            return alpha, beta, 1.
    def sf(self, x, alpha, beta, mu):
        r"""

        Survival (or reliability) function for the ExpoWeibull Distribution:

        .. math::
            R(x) = 1 - \left [ 1 - e^{-\left ( \frac{x}{\alpha} \right )^\beta} \right ]^{\mu}

        Parameters
        ----------

        x : numpy array or scalar
            The values at which the function will be calculated 
        alpha : numpy array or scalar
            scale parameter for the ExpoWeibull distribution
        beta : numpy array or scalar
            shape parameter for the ExpoWeibull distribution
        mu : numpy array or scalar
            shape parameter for the ExpoWeibull distribution

        Returns
        -------

        sf : scalar or numpy array 
            The value(s) of the reliability function at x.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import ExpoWeibull
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> ExpoWeibull.sf(x, 3, 4, 1.2)
        array([9.94911330e-01, 8.72902497e-01, 4.23286791e-01, 5.06674866e-02,
               5.34717283e-04])
        """
        return 1 - np.power(1 - np.exp(-(x / alpha)**beta), mu)

    def ff(self, x, alpha, beta, mu):
        r"""

        Failure (CDF or unreliability) function for the ExpoWeibull Distribution:

        .. math::
            F(x) = \left [ 1 - e^{-\left ( \frac{x}{\alpha} \right )^\beta} \right ]^{\mu}

        Parameters
        ----------

        x : numpy array or scalar
            The values at which the function will be calculated 
        alpha : numpy array or scalar
            scale parameter for the ExpoWeibull distribution
        beta : numpy array or scalar
            shape parameter for the ExpoWeibull distribution
        mu : numpy array or scalar
            shape parameter for the ExpoWeibull distribution

        Returns
        -------

        sf : scalar or numpy array 
            The value(s) of the failure function at x.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import ExpoWeibull
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> ExpoWeibull.ff(x, 3, 4, 1.2)
        array([0.00508867, 0.1270975 , 0.57671321, 0.94933251, 0.99946528])
        """
        return np.power(1 - np.exp(-(x / alpha)**beta), mu)

    def cs(self, x, X, alpha, beta, mu):
        r"""

        Conditional survival (or reliability) function for the ExpoWeibull Distribution:

        .. math::
            R(x, X) = \frac{R(x + X)}{R(X)}

        Parameters
        ----------

        x : numpy array or scalar
            The values at which the function will be calculated 
        alpha : numpy array or scalar
            scale parameter for the ExpoWeibull distribution
        beta : numpy array or scalar
            shape parameter for the ExpoWeibull distribution
        mu : numpy array or scalar
            shape parameter for the ExpoWeibull distribution

        Returns
        -------

        sf : scalar or numpy array 
            The value(s) of the reliability function at x.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import ExpoWeibull
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> ExpoWeibull.sf(x, 1, 3, 4, 1.2)
        array([8.77367129e-01, 4.25451775e-01, 5.09266354e-02, 5.37452200e-04,
               1.35732908e-07])
        """
        return self.sf(x + X, alpha, beta, mu) / self.sf(X, alpha, beta, mu)

    def df(self, x, alpha, beta, mu):
        r"""

        Density function for the ExpoWeibull Distribution:

        .. math::
            f(x) = \mu \left ( \frac{\beta}{\alpha} \right ) \left ( \frac{x}{\alpha} \right )^{\beta - 1} \left [ 1 - e^{-\left ( \frac{x}{\alpha} \right )^\beta} \right ]^{\mu - 1} e^{- \left ( \frac{x}{\alpha} \right )^\beta}

        Parameters
        ----------

        x : numpy array or scalar
            The values at which the function will be calculated 
        alpha : numpy array or scalar
            scale parameter for the ExpoWeibull distribution
        beta : numpy array or scalar
            shape parameter for the ExpoWeibull distribution
        mu : numpy array or scalar
            shape parameter for the ExpoWeibull distribution

        Returns
        -------

        df : scalar or numpy array 
            The value(s) of the density function at x.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import ExpoWeibull
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> ExpoWeibull.df(x, 3, 4, 1.2)
        array([0.02427515, 0.27589838, 0.53701385, 0.15943643, 0.00330058])
        """
        return (beta * mu * x**(beta - 1)) / (alpha**beta) \
                * (1 - np.exp(-(x/alpha)**beta))**(mu - 1) \
                * np.exp(-(x/alpha)**beta)

    def hf(self, x, alpha, beta, mu):
        r"""

        Instantaneous hazard rate for the ExpoWeibull Distribution:

        .. math::
            h(x) = \frac{f(x)}{R(x)}

        Parameters
        ----------

        x : numpy array or scalar
            The values at which the function will be calculated 
        alpha : numpy array or scalar
            scale parameter for the ExpoWeibull distribution
        beta : numpy array or scalar
            shape parameter for the ExpoWeibull distribution
        mu : numpy array or scalar
            shape parameter for the ExpoWeibull distribution

        Returns
        -------

        hf : scalar or numpy array 
            The value(s) of the instantaneous hazard rate at x.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import ExpoWeibull
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> ExpoWeibull.hf(x, 3, 4, 1.2)
        array([0.02439931, 0.3160701 , 1.26867613, 3.14672068, 6.17256436])
        """
        return self.df(x, alpha, beta, mu) / self.sf(x, alpha, beta, mu)

    def Hf(self, x, alpha, beta, mu):
        r"""

        Instantaneous hazard rate for the ExpoWeibull Distribution:

        .. math::
            H(x) = -\ln \left ( R(x) \right )

        Parameters
        ----------

        x : numpy array or scalar
            The values at which the function will be calculated 
        alpha : numpy array or scalar
            scale parameter for the ExpoWeibull distribution
        beta : numpy array or scalar
            shape parameter for the ExpoWeibull distribution
        mu : numpy array or scalar
            shape parameter for the ExpoWeibull distribution

        Returns
        -------

        Hf : scalar or numpy array 
            The value(s) of the cumulative hazard rate at x.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import ExpoWeibull
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> ExpoWeibull.Hf(x, 3, 4, 1.2)
        array([5.10166141e-03, 1.35931416e-01, 8.59705336e-01, 2.98247086e+00,
               7.53377239e+00])
        """
        return -np.log(self.sf(x, alpha, beta, mu))

    def qf(self, p, alpha, beta, mu):
        r"""

        Instantaneous hazard rate for the ExpoWeibull Distribution:

        .. math::
            q(p) = 

        Parameters
        ----------

        p : numpy array or scalar
            The percentiles at which the quantile will be calculated
        alpha : numpy array or scalar
            scale parameter for the ExpoWeibull distribution
        beta : numpy array or scalar
            shape parameter for the ExpoWeibull distribution
        mu : numpy array or scalar
            shape parameter for the ExpoWeibull distribution

        Returns
        -------

        Q : scalar or numpy array 
            The quantiles for the Weibull distribution at each value p

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import ExpoWeibull
        >>> p = np.array([.1, .2, .3, .4, .5])
        >>> ExpoWeibull.qf(p, 3, 4, 1.2)
        array([1.89361341, 2.2261045 , 2.46627621, 2.66992747, 2.85807988])
        """
        return alpha * (-np.log(1 - p**(1./mu)))**(1/beta)

    def mean(self, alpha, beta, mu):
        func = lambda x : x * self.df(x, alpha, beta, mu)
        top = 2 * self.qf(0.999, alpha, beta, mu)
        return integrate.quadrature(func, 0, top)[0]

    def random(self, size, alpha, beta, mu):
        U = uniform.rvs(size=size)
        return self.qf(U, alpha, beta, mu)

    def mpp_x_transform(self, x, gamma=0):
        return np.log(x - gamma)

    def mpp_y_transform(self, y, *params):
        mu = params[-1]
        mask = ((y == 0) | (y == 1))
        out = np.zeros_like(y)
        out[~mask] = np.log(-np.log((1 - y[~mask]**(1./mu))))
        out[mask] = np.nan
        return out

    def mpp_inv_y_transform(self, y, *params):
        i = len(params)
        mu = params[i-1]
        return (1 - np.exp(-np.exp(y)))**mu

    def unpack_rr(self, params, rr):
        #UPDATE ME
        if rr == 'y':
            beta  = params[0]
            alpha = np.exp(params[1]/-beta)
        elif rr == 'x':
            beta  = 1./params[0]
            alpha = np.exp(params[1] / (beta * params[0]))
        return alpha, beta, 1.
        
ExpoWeibull = ExpoWeibull_('ExpoWeibull')


from surpyval import np
from scipy.stats import uniform
from scipy.special import ndtri as z
from autograd.scipy.stats import norm
from scipy.stats import norm as scipy_norm

from surpyval import nonparametric as nonp
from surpyval import parametric as para
from surpyval.parametric.parametric_fitter import ParametricFitter

class Normal_(ParametricFitter):
    r"""

    Class used to generate the Normal (Gauss) class.

    .. code:: python

        from surpyval import Normal

    """
    def __init__(self, name):
        self.name = name
        self.k = 2
        self.bounds = ((None, None), (0, None),)
        self.support = (-np.inf, np.inf)
        self.plot_x_scale = 'linear'
        self.y_ticks = [0.001, 0.01, 0.1, 0.2, 0.3, 0.4, 
                0.5, 0.6, 0.7, 0.8, 0.9, 0.99, 0.999]
        self.param_names = ['mu', 'sigma']
        self.param_map = {
            'mu'    : 0,
            'sigma' : 1
        }

    def _parameter_initialiser(self, x, c=None, n=None, t=None):
        return para.Normal.fit(x, c, n, t, how='MPP').params

    def log_df(self, x, mu, sigma):
        return np.log(self.df(x, mu, sigma))

    def sf(self, x, mu, sigma):
        r"""

        Surival (or Reliability) function for the Normal Distribution:

        .. math::
            R(x) = 1 - \Phi \left( \frac{x - \mu}{\sigma} \right )

        Parameters
        ----------

        x : numpy array or scalar
            The values at which the function will be calculated 
        mu : numpy array or scalar
            The location parameter for the Normal distribution
        sigma : numpy array or scalar
            The scale parameter for the Normal distribution

        Returns
        -------

        sf : scalar or numpy array 
            The value(s) of the survival function at x.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Normal
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> Normal.sf(x, 3, 4)
        array([0.69146246, 0.59870633, 0.5       , 0.40129367, 0.30853754])
        """
        return norm.sf(x, mu, sigma)

    def cs(self, x, X, mu, sigma):
        r"""

        Conditional survival function for the Normal Distribution:

        .. math::
            R(x, X) = \frac{R(x + X)}{R(X)}

        Parameters
        ----------

        x : numpy array or scalar
            The value(s) at which the function will be calculated
        X : numpy array or scalar
            The value(s) at which each value(s) in x was known to have survived
        mu : numpy array or scalar
            The location parameter for the Normal distribution
        sigma : numpy array or scalar
            The scale parameter for the Normal distribution

        Returns
        -------

        cs : scalar or numpy array 
            the conditional survival probability at x

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Normal
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> Normal.cs(x, 5, 3, 4)
        array([0.73452116, 0.51421702, 0.34242113, 0.2165286 , 0.1298356 ])
        """
        return self.sf(x + X, mu, sigma) / self.sf(X, mu, sigma)

    def ff(self, x, mu, sigma):
        r"""

        CDF (or unreliability or failure) function for the Normal Distribution:

        .. math::
            F(x) = \Phi \left( \frac{x - \mu}{\sigma} \right )

        Parameters
        ----------

        x : numpy array or scalar
            The values at which the function will be calculated 
        mu : numpy array or scalar
            The location parameter for the Normal distribution
        sigma : numpy array or scalar
            The scale parameter for the Normal distribution

        Returns
        -------

        ff : scalar or numpy array 
            The value(s) of the failure function at x.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Normal
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> Normal.ff(x, 3, 4)
        array([0.30853754, 0.40129367, 0.5       , 0.59870633, 0.69146246])
        """
        return norm.cdf(x, mu, sigma)

    def df(self, x, mu, sigma):
        r"""

        Density function for the Normal Distribution:

        .. math::
            f(x) = \frac{1}{\sigma \sqrt{2\pi}}e^{-\frac{1}{2}\left ( \frac{x - \mu}{\sigma} \right )^{2}}

        Parameters
        ----------

        x : numpy array or scalar
            The values at which the function will be calculated 
        mu : numpy array or scalar
            The location parameter for the Normal distribution
        sigma : numpy array or scalar
            The scale parameter for the Normal distribution

        Returns
        -------

        df : scalar or numpy array 
            The value(s) of the density function at x.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Normal
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> Normal.df(x, 3, 4)
        array([0.08801633, 0.09666703, 0.09973557, 0.09666703, 0.08801633])
        """
        return norm.pdf(x, mu, sigma)

    def hf(self, x, mu, sigma):
        r"""

        Instantaneous hazard rate for the Normal Distribution:

        .. math::
            h(x) = \frac{\frac{1}{\sigma \sqrt{2\pi}}e^{-\frac{1}{2}\left ( \frac{x - \mu}{\sigma} \right )^{2}}}{1 - \Phi \left( \frac{x - \mu}{\sigma} \right )}

        Parameters
        ----------

        x : numpy array or scalar
            The values at which the function will be calculated 
        mu : numpy array or scalar
            The location parameter for the Normal distribution
        sigma : numpy array or scalar
            The scale parameter for the Normal distribution

        Returns
        -------

        hf : scalar or numpy array 
            The value(s) of the instantaneous hazard rate function at x.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Normal
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> Normal.hf(x, 3, 4)
        array([0.12729011, 0.16145984, 0.19947114, 0.24088849, 0.28526944])
        """
        return norm.pdf(x, mu, sigma) / self.sf(x, mu, sigma)

    def Hf(self, x, mu, sigma):
        r"""

        Cumulative hazard rate for the Normal Distribution:

        .. math::
            H(x) = -\ln \left( 1 - \Phi \left( \frac{x - \mu}{\sigma} \right ) \right )

        Parameters
        ----------

        x : numpy array or scalar
            The values at which the function will be calculated 
        mu : numpy array or scalar
            The location parameter for the Normal distribution
        sigma : numpy array or scalar
            The scale parameter for the Normal distribution

        Returns
        -------

        ff : scalar or numpy array 
            The value(s) of the cumulative hazard rate function at x.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Normal
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> Normal.Hf(x, 3, 4)
        array([0.36894642, 0.51298408, 0.69314718, 0.91306176, 1.17591176])
        """
        return -np.log(norm.sf(x, mu, sigma))

    def qf(self, p, mu, sigma):
        r"""

        Quantile function for the Normal Distribution:

        .. math::
            q(p) = \Phi^{-1} \left( p \right )

        Parameters
        ----------

        p : numpy array or scalar
            The percentiles at which the quantile will be calculated 
        mu : numpy array or scalar
            The location parameter for the Normal distribution
        sigma : numpy array or scalar
            The scale parameter for the Normal distribution

        Returns
        -------

        q : scalar or numpy array 
            The quantiles for the Normal distribution at each value p.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Normal
        >>> p = np.array([.1, .2, .3, .4, .5])
        >>> Normal.qf(p, 3, 4)
        array([-2.12620626, -0.36648493,  0.90239795,  1.98661159,  3.        ])
        """
        return scipy_norm.ppf(p, mu, sigma)

    def mean(self, mu, sigma):
        r"""

        Mean of the Normal distribution

        .. math::
            E = \mu

        Parameters
        ----------

        mu : numpy array or scalar
            The location parameter for the Normal distribution
        sigma : numpy array or scalar
            The scale parameter for the Normal distribution

        Returns
        -------

        mu : scalar or numpy array 
            The mean(s) of the Normal distribution

        Examples
        --------
        >>> from surpyval import Normal
        >>> Normal.mean(3, 4)
        3
        """
        return mu

    def moment(self, n, mu, sigma):
        r"""

        n-th (non central) moment of the Normal distribution

        .. math::
            E = ... complicated.

        Parameters
        ----------

        n : integer or numpy array of integers
            The ordinal of the moment to calculate
        mu : numpy array or scalar
            The location parameter for the Normal distribution
        sigma : numpy array or scalar
            The scale parameter for the Normal distribution

        Returns
        -------

        moment : scalar or numpy array 
            The moment(s) of the Normal distribution

        Examples
        --------
        >>> from surpyval import Normal
        >>> Normal.moment(2, 3, 4)
        25.0
        """
        return scipy_norm.moment(n, mu, sigma)

    def random(self, size, mu, sigma):
        r"""

        Draws random samples from the distribution in shape `size`

        Parameters
        ----------

        size : integer or tuple of positive integers
            Shape or size of the random draw
        mu : numpy array or scalar
            The location parameter for the Normal distribution
        sigma : numpy array or scalar
            The scale parameter for the Normal distribution

        Returns
        -------

        random : scalar or numpy array 
            Random values drawn from the distribution in shape `size`

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Normal
        >>> Normal.random(10, 3, 4)
        array([-1.28484969, -1.68138703,  0.13414348,  6.53416927, -1.95649712,
                3.09951162,  6.90469836,  4.90063467,  1.11075072,  4.97841115])
        >>> Normal.random((5, 5), 3, 4)
        array([[ 1.57569952,  4.98472487,  3.19475597,  5.12581251, -0.98020861],
               [ 6.73877217,  1.08561611,  3.07634125,  3.54656313, 13.32064634],
               [-0.45094731,  2.52588422, -1.61414841,  8.39084564, -1.35261631],
               [ 1.98090151,  8.22151826,  5.59184063, -2.62221656,  0.20879673],
               [-2.0790734 ,  2.67886095,  2.54115153,  5.49853925,  4.57056015]])
        """
        U = uniform.rvs(size=size)
        return self.qf(U, mu, sigma)

    def mpp_x_transform(self, x):
        return x

    def mpp_y_transform(self, y, *params):
        return self.qf(y, 0, 1)

    def mpp_inv_y_transform(self, y, *params):
        return self.ff(y, 0, 1)

    def unpack_rr(self, params, rr):
        if rr == 'y':
            sigma, mu = params
            mu = -mu/sigma
            sigma = 1./sigma
        elif rr == 'x':
            sigma, mu = params
        return mu, sigma

    def var_z(self, x, mu, sigma, cv_matrix):
        z_hat = (x - mu)/sigma
        var_z = (1./sigma)**2 * (cv_matrix[0, 0] + z_hat**2 * cv_matrix[1, 1] + 
            2 * z_hat * cv_matrix[0, 1])
        return var_z

    def z_cb(self, x, mu, sigma, cv_matrix, alpha_ci=0.05):
        z_hat = (x - mu)/sigma
        var_z = self.var_z(x, mu, sigma, cv_matrix)
        bounds = z_hat + np.array([1., -1.]).reshape(2, 1) * z(alpha_ci/2) * np.sqrt(var_z)
        return bounds

    # def R_cb(self, x, mu, sigma, cv_matrix, alpha_ci=0.05):
        # return self.sf(self.z_cb(x, mu, sigma, cv_matrix, alpha_ci=alpha_ci), 0, 1).T

Normal = Normal_('Normal')
Gauss = Normal_('Gauss')



"""
This code was created for and sponsored by Cartiga (www.cartiga.com). 
Cartiga makes no representations or warranties in connection with the code 
and waives any and all liability in connection therewith. Your use of the 
code constitutes acceptance of these terms.

Copyright 2022 Cartiga LLC
"""

import pytest
import numpy as np
from ..datasets import (
    Lung,
    Rossi,
    Heart
)
from surpyval.regression import CoxPH

def test_coxph_against_ll_rossi_static():
    ll_answer = np.array([-0.37942216, -0.05743772,  0.31389978,
                          -0.14979572, -0.43370385, -0.08487107, 
                           0.09149708])

    rossi = Rossi.data
    model = CoxPH.fit_from_df(rossi, 
            x_col='week',
            c_col='arrest',
            Z_cols=['fin', 'age', 'race', 'wexp', 'mar', 'paro', 'prio'],
            method="efron")

    assert np.allclose(model.beta, ll_answer)

# Examples taken from:
# http://www.sthda.com/english/wiki/cox-proportional-hazards-model

def test_coxph_against_r_lung_1():
    r_answer = np.array([-0.5310235])

    lung = Lung.data
    x = lung['time'].values
    c = lung['status'].values
    Z = lung[['sex']].values

    model = CoxPH.fit(x=x, Z=Z, c=c, method="efron")

    assert np.allclose(model.beta, r_answer)

def test_coxph_against_r_lung_2():
    r_answer = np.array([0.01106676, -0.55261240, 0.46372848])

    lung = Lung.data
    model = CoxPH.fit_from_df(
        lung, 
        x_col='time',
        c_col='status',
        Z_cols=['age', 'sex', 'ph.ecog'],
        method="efron",
    )

    assert np.allclose(model.beta, r_answer)



import numpy as np
from surpyval.nonparametric.nonparametric_fitter import NonParametricFitter


def kaplan_meier(r, d):
    R = 1 - (d / r)
    R[np.isnan(R)] = 0
    old_err_state = np.seterr(under='raise')

    try:
        R = np.cumprod(R)
    except FloatingPointError:
        R = np.cumsum(np.log(1 - (d / r)))
        R = np.exp(R)

    np.seterr(**old_err_state)
    return R


class KaplanMeier_(NonParametricFitter):
    r"""
    Kaplan-Meier estimator class. Calculates the Non-Parametric
    estimate of the survival function using:

    .. math::
        R(x) = \prod_{i:x_{i} \leq x}^{}
            \left ( 1 - \frac{d_{i} }{r_{i}}  \right )

    Examples
    --------
    >>> import numpy as np
    >>> from surpyval import KaplanMeier
    >>> x = np.array([1, 2, 3, 4, 5])
    >>> model = KaplanMeier.fit(x)
    >>> model.R
    array([0.8, 0.6, 0.4, 0.2, 0. ])
    """
    def __init__(self):
        self.how = 'Kaplan-Meier'


KaplanMeier = KaplanMeier_()



import pytest
import numpy as np
import lifelines
import surpyval as surv

from collections import namedtuple
from lifelines.datasets import *
from reliability.Fitters import *

# Datasets in x, c, n: as namedtuples
SurvivalData = namedtuple('SurvivalData', ['x', 'c', 'n', 'name'])
IntervalSurvivalData = namedtuple('IntervalSurvivalData',
                                  ['left', 'right', 'name'])

# Canadian Senators
df = load_canadian_senators()
x = df['diff_days'].values
c = 1 - df['observed'].astype(int)
zero_idx = (x == 0)
x = x[~zero_idx]
c = c[~zero_idx]
n = np.ones_like(x)
canadian_senators = SurvivalData(x, c, n, 'canadian_senators')

# Bacteria...
df = load_c_botulinum_lag_phase()
left = df['lower_bound_days']
right = df['upper_bound_days']
bacteria = IntervalSurvivalData(left, right, 'bacteria')

# Political Durations
df = load_dd()
x = df['duration'].values
c = 1 - df['observed'].astype(int)
n = np.ones_like(x)
politics = SurvivalData(x, c, n, 'politics')

df = load_diabetes()
left = df['left']
right = df['right']
diabetes = IntervalSurvivalData(left, right, 'diabetes')

# ???
df = load_g3()
x = df['time'].values
c = 1 - df['event'].astype(int)
n = np.ones_like(x)
g3 = SurvivalData(x, c, n, 'g3')

# ???
df = load_gbsg2()
x = df['time'].values
c = 1 - df['cens'].astype(int)
n = np.ones_like(x)
gbsg2 = SurvivalData(x, c, n, 'gbsg2')

# holl_molly_polly...??!!!?!?!?!?!
df = load_holly_molly_polly()
x = df['T'].values
c = np.zeros_like(x)
n = np.ones_like(x)
holly_molly_polly = SurvivalData(x, c, n, 'holly_molly_polly')

# kidney
df = load_kidney_transplant()
x = df['time'].values
c = 1 - df['death'].astype(int)
n = np.ones_like(x)
kidney = SurvivalData(x, c, n, 'kidney')

# larynx
df = load_larynx()
x = df['time'].values
c = np.zeros_like(x)
n = np.ones_like(x)
larynx = SurvivalData(x, c, n, 'larynx')

# leukemia
df = load_leukemia()
x = df['t'].values
c = 1 - df['status'].astype(int)
n = np.ones_like(x)
leukemia = SurvivalData(x, c, n, 'leukemia')

# lung
df = load_lung()
x = df['time'].dropna()
c = np.zeros_like(x)
n = np.ones_like(x)
lung = SurvivalData(x, c, n, 'lung')

# lupus
df = load_lupus()
x = df['time_elapsed_between_estimated_onset_and_diagnosis_(months)'].dropna()
c = 1 - df['dead'].astype(int)
n = np.ones_like(x)
lupus = SurvivalData(x, c, n, 'lupus')

# lymph
df = load_lymph_node()
x = df['survtime'].dropna()
c = 1 - df['censdead'].astype(int)
n = np.ones_like(x)
lymph = SurvivalData(x, c, n, 'lymph')

# lymphoma
df = load_lymphoma()
x = df['Time'].dropna()
c = 1 - df['Censor'].astype(int)
n = np.ones_like(x)
lymphoma = SurvivalData(x, c, n, 'lymphoma')

# mice
df = load_mice()
left = df['l']
right = df['u']
mice = IntervalSurvivalData(left, right, 'mice')

# aids
df = load_multicenter_aids_cohort_study()
x = df['T'].dropna()
c = 1 - df['D'].astype(int)
n = np.ones_like(x)
aids = SurvivalData(x, c, n, 'aids')

# nh4
df = load_nh4()
x = df['Week'].dropna()
c = 1 - df['Censored'].astype(int)
n = np.ones_like(x)
nh4 = SurvivalData(x, c, n, 'nh4')

# panel
df = load_panel_test()
x = df['t'].dropna()
c = 1 - df['E'].astype(int)
n = np.ones_like(x)
panel = SurvivalData(x, c, n, 'panel')

# recur
df = load_recur()
x = df['AGE'].dropna()
c = 1 - df['CENSOR'].astype(int)
n = np.ones_like(x)
recur = SurvivalData(x, c, n, 'recur')

# reg
df = load_regression_dataset()
x = df['T'].dropna()
c = 1 - df['E'].astype(int)
n = np.ones_like(x)
reg = SurvivalData(x, c, n, 'reg')

# rossi
df = load_rossi()
x = df['week'].dropna()
c = 1 - df['arrest'].astype(int)
n = np.ones_like(x)
rossi = SurvivalData(x, c, n, 'rossi')

# static
df = load_static_test()
x = df['t'].dropna() + 1e-10
c = 1 - df['E'].astype(int)
n = np.ones_like(x)
static = SurvivalData(x, c, n, 'static')

# walton
df = load_waltons()
x = df['T'].dropna()
c = 1 - df['E'].astype(int)
n = np.ones_like(x)
walton = SurvivalData(x, c, n, 'walton')

def id_func(val):
    if isinstance(val, SurvivalData):
        return val.name
    elif isinstance(val, IntervalSurvivalData):
        return val.name

xcn_datasets = [
    canadian_senators,
    politics,
    g3,
    gbsg2,
    holly_molly_polly,
    kidney,
    larynx,
    leukemia,
    lung,
    lupus,
    lymph,
    lymphoma,
    aids,
    nh4,
    panel,
    recur,
    reg,
    rossi,
    # static,
    walton
]

int_datasets = [
    bacteria,
    diabetes,
    mice
]

wf = lifelines.WeibullFitter()
lnf = lifelines.LogNormalFitter()
llf = lifelines.LogLogisticFitter()
ef = lifelines.ExponentialFitter()

DISTS = {
    'Weibull': (wf, surv.Weibull),
    'Exponential': (ef, surv.Exponential),
    'LogNormal': (lnf, surv.LogNormal),
    'LogLogistic': (llf, surv.LogLogistic)
}

REL_DISTS = {
    'Exponential': (Fit_Exponential_1P, surv.Exponential),
    'Weibull': (Fit_Weibull_2P, surv.Weibull),
    'Gamma': (Fit_Gamma_2P, surv.Gamma),
    'LogNormal': (Fit_Lognormal_2P, surv.LogNormal),
    'LogLogistic': (Fit_Loglogistic_2P, surv.LogLogistic),
    'Normal': (Fit_Normal_2P, surv.Normal),
    'Gumbel': (Fit_Gumbel_2P, surv.Gumbel),
    'Beta': (Fit_Beta_2P, surv.Beta),
}   

def generate_case():
    for i, data in enumerate(xcn_datasets):
        yield data


def generate_real_cases():
    for dist in DISTS.keys():
        for data in xcn_datasets:
            yield data, dist

def generate_real_cases_reliability():
    for dist in REL_DISTS.keys():
        for data in xcn_datasets:
            yield data, dist

def generate_real_cases_int():
    for dist in DISTS.keys():
        for data in int_datasets:
            yield data, dist

def params_with_xcn_data_rel(data, surpyval_fitter, rel_fitter):
    if surpyval_fitter.name == 'Beta':
        x = data.x/(data.x.max()+1)
    else:
        x = data.x
    f, s = surv.xcn_to_fs(x, data.c, data.n)
    if s == []:
        s = None
    rel_model = rel_fitter(f, s)
    if surpyval_fitter.name == 'Exponential':
        rel_params = rel_model.Lambda
    elif surpyval_fitter.name in ['Weibull', 'Gamma',
                                  'LogLogistic', 'Beta']:
        rel_params = np.array([rel_model.alpha, rel_model.beta])
    elif surpyval_fitter.name in ['LogNormal', 'Normal', 'Gumbel']:
        rel_params = np.array([rel_model.mu, rel_model.sigma])

    surp_est = surpyval_fitter.fit(x,
                                   data.c,
                                   data.n)
    if np.allclose(rel_params, surp_est.params, 1e-1):
        return True
    else:
        # reliability has performance that is to be desired. So check that
        # loglike is better or within a small tolerance:
        return (surp_est.neg_ll() - (-rel_model.loglik)) < 1e-5

def params_with_xcn_data(data, surpyval_fitter, lifelines_fitter):
    ll_est = lifelines_fitter.fit(data.x,
                                  1 - data.c,
                                  weights=data.n).params_.values
    surp_est = surpyval_fitter.fit(data.x,
                                   data.c,
                                   data.n).params
    if surpyval_fitter.name == 'Exponential':
        surp_est = 1./surp_est
    return ll_est, surp_est

def params_with_int_data(data, surpyval_fitter, lifelines_fitter):
    ll_est = lifelines_fitter.fit_interval_censoring(data.left,
                                                     data.right).params_.values
    surp_est = surpyval_fitter.fit(xl=data.left,
                                   xr=data.right).params
    if surpyval_fitter.name == 'Exponential':
        surp_est = 1./surp_est
    return ll_est, surp_est

@pytest.mark.parametrize("data",
                         generate_case(),
                         ids=id_func)
def test_weibull_offset_with_real(data):
    # Known issues - distribution too far off being Weibull to work with offset
    if data.name in ['gbsg2',
                     'kidney',
                     'lymph',
                     'aids']:
        assert True
    else:
        surpyval_fitter = surv.Weibull
        fitted = surpyval_fitter.fit(data.x,
                                     data.c,
                                     data.n,
                                     offset=True)
        assert fitted.res.success or ('Desired error' in fitted.res.message)

@pytest.mark.parametrize("data,dist",
                         generate_real_cases(),
                         ids=id_func)
def test_against_lifelines_with_real_data(data, dist):
    ll_fitter = DISTS[dist][0]
    surp_fitter = DISTS[dist][1]
    assert np.allclose(*params_with_xcn_data(data,
                                             surp_fitter,
                                             ll_fitter), 1e-1)

@pytest.mark.parametrize("data,dist",
                         generate_real_cases_int(),
                         ids=id_func)
def test_against_lifelines_with_real_data_interval(data, dist):
    ll_fitter = DISTS[dist][0]
    surp_fitter = DISTS[dist][1]
    assert np.allclose(*params_with_int_data(data,
                                             surp_fitter,
                                             ll_fitter), 1e-1)

@pytest.mark.parametrize("data,dist",
                         generate_real_cases_reliability(),
                         ids=id_func)
def test_against_reliability_with_real_data(data, dist):
    rel_fitter = REL_DISTS[dist][0]
    surp_fitter = REL_DISTS[dist][1]
    assert params_with_xcn_data_rel(data, surp_fitter, rel_fitter)



import numpy as np
from scipy.stats import t, norm
from scipy.interpolate import interp1d

import matplotlib.pyplot as plt
import pandas as pd


def interp_function(x, y, kind):
    return interp1d(x, y, kind=kind,
                    bounds_error=False,
                    fill_value=np.nan)


class NonParametric():
    """
    Result of ``.fit()`` method for every non-parametric
    surpyval distribution. This means that each of the
    methods in this class can be called with a model created
    from the ``NelsonAalen``, ``KaplanMeier``,
    ``FlemingHarrington``, or ``Turnbull`` estimators.
    """
    def __repr__(self):
        out = ('Non-Parametric SurPyval Model'
               + '\n============================='
               + '\nModel            : {dist}'
               ).format(dist=self.model)

        if hasattr(self, 'data'):
            if 'estimator' in self.data:
                out += '\nEstimator        : {turnbull}'.format(
                    turnbull=self.data['estimator'])

        return out

    def sf(self, x, interp='step'):
        r"""

        Surival (or Reliability) function with the
        non-parametric estimates from the data.

        Parameters
        ----------

        x : array like or scalar
            The values of the random variables at which t
            he survival function will be calculated.

        Returns
        -------

        sf : scalar or numpy array
            The value(s) of the survival function at each x


        Examples
        --------
        >>> from surpyval import NelsonAalen
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> model = NelsonAalen.fit(x)
        >>> model.sf(2)
        array([0.63762815])
        >>> model.sf([1., 1.5, 2., 2.5])
        array([0.81873075, 0.81873075, 0.63762815, 0.63762815])
        """
        x = np.atleast_1d(x)
        idx = np.argsort(x)
        rev = np.argsort(idx)
        x = x[idx]
        if interp == 'step':
            idx = np.searchsorted(self.x, x, side='right') - 1
            R = self.R[idx]
            R = np.where(idx < 0, 1, R)
            R = np.where(np.isposinf(x), 0, R)
        else:
            R = interp_function(self.x, self.R, kind=interp)(x)

        R = R[rev]
        # Maybe set a parameter where 'extrapolate' is False
        # x = x[rev]
        # R = np.where(x < self.x.min(), np.nan, R)
        # R = np.where(x > self.x.max(), np.nan, R)
        return R

    def ff(self, x, interp='step'):
        r"""

        CDF (failure or unreliability) function with the
        non-parametric estimates from the data

        Parameters
        ----------

        x : array like or scalar
            The values of the random variables at which
            the survival function will be calculated.

        Returns
        -------

        ff : scalar or numpy array
            The value(s) of the failure function at each x


        Examples
        --------
        >>> from surpyval import NelsonAalen
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> model = NelsonAalen.fit(x)
        >>> model.ff(2)
        array([0.36237185])
        >>> model.ff([1., 1.5, 2., 2.5])
        array([0.18126925, 0.18126925, 0.36237185, 0.36237185])
        """
        return 1 - self.sf(x, interp=interp)

    def hf(self, x, interp='step'):
        r"""

        Instantaneous hazard function with the non-parametric
        estimates from the data. This is calculated using simply
        the difference between consecutive H(x).

        Parameters
        ----------

        x : array like or scalar
            The values of the random variables at which
            the survival function will be calculated

        Returns
        -------

        hf : scalar or numpy array
            The value(s) of the failure function at each x


        Examples
        --------
        >>> from surpyval import NelsonAalen
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> model = NelsonAalen.fit(x)
        >>> model.ff(2)
        array([0.36237185])
        >>> model.ff([1., 1.5, 2., 2.5])
        array([0.18126925, 0.18126925, 0.36237185, 0.36237185])
        """
        idx = np.argsort(x)
        rev = np.argsort(idx)
        x = x[idx]
        hf = np.diff(np.hstack([self.Hf(x[0], interp=interp),
                                self.Hf(x, interp=interp)]))
        hf[0] = hf[1]
        hf = pd.Series(hf)
        hf[hf == 0] = np.nan
        hf = hf.ffill().values
        return hf[rev]

    def df(self, x, interp='step'):
        r"""

        Density function with the non-parametric estimates
        from the data. This is calculated using the relationship
        between the hazard function and the density:

        .. math::
            f(x) = h(x)e^{-H(x)}

        Parameters
        ----------

        x : array like or scalar
            The values of the random variables at which the
            survival function will be calculated

        Returns
        -------

        df : scalar or numpy array
            The value(s) of the density function at x


        Examples
        --------
        >>> from surpyval import NelsonAalen
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> model = NelsonAalen.fit(x)
        >>> model.df(2)
        array([0.28693267])
        >>> model.df([1., 1.5, 2., 2.5])
        array([0.16374615, 0.        , 0.15940704, 0.        ])
        """
        return self.hf(x, interp=interp) * np.exp(-self.Hf(x, interp=interp))

    def Hf(self, x, interp='step'):
        r"""

        Cumulative hazard rate with the non-parametric estimates
        from the data. This is calculated using the relationship
        between the hazard function and the density:

        .. math::
            H(x) = -\ln (R(x))

        Parameters
        ----------

        x : array like or scalar
            The values of the random variables at which the
            function will be calculated.

        Returns
        -------

        Hf : scalar or numpy array
            The value(s) of the density function at x

        Examples
        --------
        >>> from surpyval import NelsonAalen
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> model = NelsonAalen.fit(x)
        >>> model.Hf(2)
        array([0.45])
        >>> model.df([1., 1.5, 2., 2.5])
        model.Hf([1., 1.5, 2., 2.5])
        """
        return -np.log(self.sf(x, interp=interp))

    def cb(self, x, on='sf', bound='two-sided', interp='step',
           alpha_ci=0.05, bound_type='exp', dist='z'):
        r"""

        Confidence bounds of the ``on`` function at the
        ``alpa_ci`` level of significance. Can be the upper,
        lower, or two-sided confidence by changing value of ``bound``.
        Can change the bound type to be regular or exponential
        using either the 't' or 'z' statistic.

        Parameters
        ----------

        x : array like or scalar
            The values of the random variables at which the confidence bounds
            will be calculated
        on : ('sf', 'ff', 'Hf'), optional
            The function on which the confidence bound will be calculated.
        bound : ('two-sided', 'upper', 'lower'), str, optional
            Compute either the two-sided, upper or lower confidence bound(s).
            Defaults to two-sided.
        interp : ('step', 'linear', 'cubic'), optional
            How to interpolate the values between observations. Survival
            statistics traditionally uses step functions, but can use
            interpolated values if desired. Defaults to step.
        alpha_ci : scalar, optional
            The level of significance at which the bound will be computed.
        bound_type : ('exp', 'regular'), str, optional
            The method with which the bounds will be calculated. Using regular
            will allow for the bounds to exceed 1 or be less than 0. Defaults
            to exp as this ensures the bounds are within 0 and 1.
        dist : ('t', 'z'), str, optional
            The statistic to use in calculating the bounds (student-t or
            normal). Defaults to the normal (z).

        Returns
        -------

        cb : scalar or numpy array
            The value(s) of the upper, lower, or both confidence bound(s) of
            the selected function at x

        Examples
        --------
        >>> from surpyval import NelsonAalen
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> model = NelsonAalen.fit(x)
        >>> model.cb([1., 1.5, 2., 2.5], bound='lower', dist='t')
        array([0.11434813, 0.11434813, 0.04794404, 0.04794404])
        >>> model.cb([1., 1.5, 2., 2.5])
        array([[0.97789387, 0.16706394],
               [0.97789387, 0.16706394],
               [0.91235117, 0.10996882],
               [0.91235117, 0.10996882]])

        References
        ----------

        http://reliawiki.org/index.php/Non-Parametric_Life_Data_Analysis

        """
        if on in ['df', 'hf']:
            raise ValueError("NonParametric cannot do confidence bounds on "
                             + "density or hazard rate functions. Try Hf, "
                             + "ff, or sf")

        old_err_state = np.seterr(all='ignore')

        # Reverse for ff and F
        if on in ['ff', 'F', 'Hf', 'hf', 'df'] and bound == 'lower':
            bound = 'upper'
        elif on in ['ff', 'F', 'Hf', 'hf', 'df'] and bound == 'upper':
            bound = 'lower'

        cb = self.R_cb(x,
                       bound=bound,
                       interp=interp,
                       alpha_ci=alpha_ci,
                       bound_type=bound_type,
                       dist=dist
                       )

        if (on == 'ff') or (on == 'F'):
            cb = 1. - cb

        elif on == 'Hf':
            cb = -np.log(cb)

        elif (on == 'sf') or (on == 'R'):
            if bound == 'two-sided':
                cb = np.fliplr(cb)

        np.seterr(**old_err_state)

        return cb

    def R_cb(self, x, bound='two-sided', interp='step', alpha_ci=0.05,
             bound_type='exp', dist='z'):

        if bound_type not in ['exp', 'normal']:
            return ValueError("'bound_type' must be in ['exp', 'normal']")
        if dist not in ['t', 'z']:
            return ValueError("'dist' must be in ['t', 'z']")

        confidence = 1. - alpha_ci

        old_err_state = np.seterr(all='ignore')

        x = np.atleast_1d(x)
        if bound in ['upper', 'lower']:
            if dist == 't':
                stat = t.ppf(1 - confidence, self.r - 1)
            else:
                stat = norm.ppf(1 - confidence, 0, 1)
            if bound == 'upper':
                stat = -stat
        elif bound == 'two-sided':
            if dist == 't':
                stat = t.ppf((1 - confidence) / 2, self.r - 1)
            else:
                stat = norm.ppf((1 - confidence) / 2, 0, 1)
            stat = np.array([-1, 1]).reshape(2, 1) * stat

        if bound_type == 'exp':
            # Exponential Greenwood confidence
            # print(self.greenwood)
            R_out = self.greenwood * 1. / (np.log(self.R)**2)
            R_out = np.log(-np.log(self.R)) - stat * np.sqrt(R_out)
            R_out = np.exp(-np.exp(R_out))
            R_out = np.where(self.greenwood == 0, 1, R_out)
        else:
            # Normal Greenwood confidence
            R_out = self.R + np.sqrt(self.greenwood * self.R**2) * stat

        # Allows for confidence bound to be estimated up to the last value.
        # only used in event that there is no right censoring.
        if bound == 'upper':
            R_out = np.where(np.isfinite(R_out), R_out, np.nanmin(R_out))
        elif bound == 'lower':
            R_out = np.where(np.isfinite(R_out), R_out, 0)
        else:
            R_out[0, :] = np.where(np.isfinite(R_out[0, :]),
                                   R_out[0, :],
                                   np.nanmin(R_out[0, :]))
            R_out[1, :] = np.where(np.isfinite(R_out[1, :]), R_out[1, :], 0)

        if interp == 'step':
            idx = np.searchsorted(self.x, x, side='right') - 1
            if bound == 'two-sided':
                R_out = R_out[:, idx]
                R_out = np.where(idx < 0, 1, R_out)
            else:
                R_out = R_out[idx]
                R_out = np.where(idx < 0, 1, R_out)

        else:
            if bound == 'two-sided':
                R1 = interp_function(self.x, R_out[0, :], kind=interp)(x)
                R2 = interp_function(self.x, R_out[1, :], kind=interp)(x)
                R_out = np.vstack([R1, R2])
            else:
                R_out = interp_function(self.x, R_out, kind=interp)(x)

        # The question remains. Should bounds above and below observed values
        # be calculable?...
        mask = (x < self.x.min()) | (x > self.x.max())
        R_out = np.where(mask, np.nan, R_out)

        if bound == 'two-sided':
            R_out = R_out.T

        np.seterr(**old_err_state)

        return R_out

    def random(self, size):
        return np.random.choice(self.x, size=size)

    def get_plot_data(self, **kwargs):
        y_scale_min = 0
        y_scale_max = 1

        # x-axis
        x_min = 0
        x_max = np.max(self.x)

        diff = (x_max - x_min) / 10
        x_scale_min = x_min
        x_scale_max = x_max + diff

        cbs = self.R_cb(self.x, **kwargs)

        return {
            'x_scale_min': x_scale_min,
            'x_scale_max': x_scale_max,
            'y_scale_min': y_scale_min,
            'y_scale_max': y_scale_max,
            'cbs': cbs,
            'x_': self.x,
            'R': self.R,
            'F': self.F
        }

    def plot(self, ax=None, **kwargs):
        r"""
        Creates a plot of the survival function.
        """
        if ax is None:
            ax = plt.gcf().gca()

        plot_bounds = kwargs.pop('plot_bounds', True)
        interp = kwargs.pop('interp', 'step')
        bound = kwargs.pop('bound', 'two-sided')
        alpha_ci = kwargs.pop('alpha_ci', 0.05)
        bound_type = kwargs.pop('bound_type', 'exp')
        dist = kwargs.pop('dist', 'z')

        d = self.get_plot_data(interp=interp,
                               bound=bound,
                               alpha_ci=alpha_ci,
                               bound_type=bound_type,
                               dist=dist)
        # MAKE THE PLOT
        # Set the y limits
        ax.set_ylim([d['y_scale_min'], d['y_scale_max']])

        # Label it
        ax.set_title('Model Survival Plot')
        ax.set_ylabel('R')
        if interp != 'step':
            ax.plot(d['x_'], d['R'])
            if plot_bounds:
                ax.plot(d['x_'], d['cbs'], color='r')
        else:
            ax.step(d['x_'], d['R'], where='post')
            if plot_bounds:
                ax.step(d['x_'], d['cbs'], color='r', where='post')

        return ax



from surpyval import np
from scipy.stats import uniform
from scipy.special import ndtri as z

from surpyval import xcn_handler
from surpyval.parametric.parametric_fitter import ParametricFitter

class Logistic_(ParametricFitter):
    def __init__(self, name):
        self.name = name
        self.k = 2
        self.bounds = ((None, None), (0, None),)
        self.support = (-np.inf, np.inf)
        self.plot_x_scale = 'linear'
        self.y_ticks = [0.0001, 0.0002, 0.0003, 0.001, 0.002, 
            0.003, 0.005, 0.01, 0.02, 0.03, 0.05, 
            0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 
            0.9, 0.95, 0.99, 0.999, 0.9999]
        self.param_names = ['mu', 'sigma']
        self.param_map = {
            'mu'    : 0,
            'sigma' : 1
        }

    def _parameter_initialiser(self, x, c=None, n=None, t=None, offset=False):
        return self.fit(x, c, n, t, how='MPP').params
        x, c, n = xcn_handler(x, c, n)
        flag = (c == 0).astype(int)
        if offset:
            return x.sum() / (n * flag).sum(), 1., 1.
        else:
            return x.sum() / (n * flag).sum(), 1.

    def sf(self, x, mu, sigma):
        r"""

        Survival (or reliability) function for the Logistic Distribution:

        .. math::
            R(x) = 1 - \frac{1}{1 + e^{- \left ( x - \mu \right )/ \sigma}}

        Parameters
        ----------

        x : numpy array or scalar
            The values at which the function will be calculated 
        mu : numpy array or scalar
            The location parameter for the Logistic distribution
        sigma : numpy array or scalar
            The scale parameter for the Logistic distribution

        Returns
        -------

        sf : scalar or numpy array 
            The value(s) of the reliability function at x.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Logistic
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> Logistic.sf(x, 3, 4)
        array([0.62245933, 0.5621765 , 0.5       , 0.4378235 , 0.37754067])
        """
        return 1 - self.ff(x, mu, sigma)

    def cs(self, x, X, mu, sigma):
        return self.sf(x + X, mu, sigma) / self.sf(X, mu, sigma)

    def ff(self, x, mu, sigma):
        r"""

        Failure (CDF or unreliability) function for the Logistic Distribution:

        .. math::
            F(x) = \frac{1}{1 + e^{- \left ( x - \mu \right )/ \sigma}}

        Parameters
        ----------

        x : numpy array or scalar
            The values at which the function will be calculated 
        mu : numpy array or scalar
            The location parameter for the Logistic distribution
        sigma : numpy array or scalar
            The scale parameter for the Logistic distribution

        Returns
        -------

        ff : scalar or numpy array 
            The value(s) of the failure function at x.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Logistic
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> Logistic.ff(x, 3, 4)
        array([0.37754067, 0.4378235 , 0.5       , 0.5621765 , 0.62245933])
        """
        z = (x - mu) / sigma
        return 1. / (1 + np.exp(-z))

    def df(self, x, mu, sigma):
        r"""

        Failure (CDF or unreliability) function for the Logistic Distribution:

        .. math::
            f(x) = \frac{e^{-\left ( x - \mu \right ) / \sigma}}{\sigma \left ( 1 + e^{-\left ( x - \mu \right )/ \sigma}\right )^2}

        Parameters
        ----------

        x : numpy array or scalar
            The values at which the function will be calculated 
        mu : numpy array or scalar
            The location parameter for the Logistic distribution
        sigma : numpy array or scalar
            The scale parameter for the Logistic distribution

        Returns
        -------

        df : scalar or numpy array 
            The value(s) of the density function at x.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Logistic
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> Logistic.df(x, 3, 4)
        array([0.05875093, 0.06153352, 0.0625    , 0.06153352, 0.05875093])
        """
        z = (x - mu) / sigma
        return np.exp(-z) / (sigma * (1 + np.exp(-z))**2)

    def hf(self, x, mu, sigma):
        r"""

        Instantaneous hazard rate for the Logistic Distribution:

        .. math::
            h(x) = \frac{f(x)}{R(x)}

        Parameters
        ----------

        x : numpy array or scalar
            The values at which the function will be calculated 
        mu : numpy array or scalar
            The location parameter for the Logistic distribution
        sigma : numpy array or scalar
            The scale parameter for the Logistic distribution

        Returns
        -------

        hf : scalar or numpy array 
            The value(s) of the instantaneous hazard rate at x.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Logistic
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> Logistic.hf(x, 3, 4)
        array([0.09438517, 0.10945587, 0.125     , 0.14054413, 0.15561483])
        """
        return self.df(x, mu, sigma) / self.sf(x, mu, sigma)

    def Hf(self, x, mu, sigma):
        r"""

        Cumulative hazard rate for the Logistic distribution:

        .. math::
            h(x) = -\ln \left ( R(x) \right )

        Parameters
        ----------

        x : numpy array or scalar
            The values at which the function will be calculated 
        mu : numpy array or scalar
            The location parameter for the Logistic distribution
        sigma : numpy array or scalar
            The scale parameter for the Logistic distribution

        Returns
        -------

        hf : scalar or numpy array 
            The value(s) of the cumulative hazard rate at x.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Logistic
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> Logistic.Hf(x, 3, 4)
        array([0.47407698, 0.57593942, 0.69314718, 0.82593942, 0.97407698])
        """
        return -np.log(self.sf(x, mu, sigma))

    def qf(self, p, mu, sigma):
        r"""

        Quantile function for the Logistic distribution:

        .. math::
            q(p) = \mu + \sigma \ln \left ( \frac{p}{1 - p} \right)

        Parameters
        ----------

        p : numpy array or scalar
            The percentiles at which the quantile will be calculated
        mu : numpy array or scalar
            The location parameter for the Logistic distribution
        sigma : numpy array or scalar
            The scale parameter for the Logistic distribution

        Returns
        -------

        q : scalar or numpy array 
            The quantiles for the Logistic distribution at each value p

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Logistic
        >>> p = np.array([.1, .2, .3, .4, .5])
        >>> Logistic.qf(p, 3, 4)
        array([-5.78889831, -2.54517744, -0.38919144,  1.37813957,  3.        ])
        """
        return mu + sigma * np.log(p/(1 - p))

    def mean(self, mu, sigma):
        r"""

        Mean of the Logistic distribution

        .. math::
            E = \mu

        Parameters
        ----------

        mu : numpy array or scalar
            The location parameter for the Logistic distribution
        sigma : numpy array or scalar
            The scale parameter for the Logistic distribution

        Returns
        -------

        mu : scalar or numpy array 
            The mean(s) of the Logistic distribution

        Examples
        --------
        >>> from surpyval import Logistic
        >>> Logistic.mean(3, 4)
        3
        """
        return mu

    def random(self, size, mu, sigma):
        r"""

        Draws random samples from the distribution in shape `size`

        Parameters
        ----------

        size : integer or tuple of positive integers
            Shape or size of the random draw
        mu : numpy array or scalar
            The location parameter for the Logistic distribution
        sigma : numpy array or scalar
            The scale parameter for the Logistic distribution

        Returns
        -------

        random : scalar or numpy array 
            Random values drawn from the distribution in shape `size`

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Logistic
        >>> Logistic.random(10, 3, 4)
        array([-8.03085073, -1.69001847,  5.25971637,  4.49119392,  3.92027233,
               -0.8320818 , -7.08778338,  5.01180405,  0.82373259,  8.51506487])
        >>> Logistic.random((5, 5), 3, 4)
        array([[ 7.11691946, 14.31662627,  8.5383889 ,  1.26608344,  0.97633704],
               [-7.11229405,  8.56748118,  1.5959416 , -3.89229554, -2.44623464],
               [ 5.58805039, -0.11768336, -0.55000158,  8.5302643 ,  6.92591024],
               [-2.88281091, -9.79724128, -3.80713019,  1.74120972, 15.37924263],
               [-4.42521443, -0.69577732,  3.54658395,  2.82310964,  3.95850831]])
        """
        U = uniform.rvs(size=size)
        return self.qf(U, mu, sigma)

    def mpp_x_transform(self, x, gamma=0):
        return x - gamma

    def mpp_y_transform(self, y, *params):
        mask = ((y == 0) | (y == 1))
        out = np.zeros_like(y)
        out[~mask] = -np.log(1./y[~mask] - 1)
        out[mask] = np.nan
        return out

    def mpp_inv_y_transform(self, y, *params):
        return 1./(np.exp(-y) + 1)

    def unpack_rr(self, params, rr):
        if   rr == 'y':
            sigma = 1./params[0]
            mu    = -sigma * params[1]
        elif rr == 'x':
            sigma  = params[0]
            mu = params[1]
        return mu, sigma

Logistic = Logistic_('Logistic')


from surpyval import np
from autograd import jacobian, hessian
from autograd.numpy.linalg import inv
from scipy.optimize import minimize, approx_fprime
import sys
import copy


def _create_censor_flags(x_mle, gamma, c, dist):
    if 2 in c:
        l_flag = x_mle[:, 0] <= np.min([gamma, dist.support[0]])
        r_flag = x_mle[:, 1] >= dist.support[1]
        mask = np.vstack([l_flag, r_flag]).T
        inf_c_flags = (mask).astype(int)
        x_mle[(c == 2).reshape(-1, 1) & mask] = 1
    else:
        inf_c_flags = np.zeros_like(x_mle)

    return inf_c_flags, x_mle


def mle(model):
    """
    Maximum Likelihood Estimation (MLE)

    """
    dist = model.dist
    x, c, n, t = (model.data['x'], model.data['c'],
                  model.data['n'], model.data['t'])
    const = model.fitting_info['const']
    trans = model.fitting_info['transform']
    inv_trans = model.fitting_info['inv_trans']
    init = model.fitting_info['init']
    fixed_idx = model.fitting_info['fixed_idx']
    offset = model.offset
    lfp = model.lfp
    zi = model.zi

    if hasattr(dist, 'mle'):
        return dist.mle(x, c, n, t, const, trans,
                        inv_trans, init, fixed_idx, offset)

    results = {}

    """
    Need to flag entries where truncation is inf or -inf so that the autograd
    doesn't fail. Because autograd fails if it encounters any inf, nan, -inf
    etc even if they don't affect the gradient. A must for autograd
    """
    t_flags = np.ones_like(t)
    t_mle = copy.copy(t)
    # Create flags to indicate where the truncation values are infinite
    t_flags[:, 0] = np.where(np.isfinite(t[:, 0]), 1, 0)
    t_flags[:, 1] = np.where(np.isfinite(t[:, 1]), 1, 0)
    # Convert the infinite values to a finite value to ensure
    # the autodiff functions don't fail
    t_mle[:, 0] = np.where(t_flags[:, 0] == 1, t[:, 0], 1)
    t_mle[:, 1] = np.where(t_flags[:, 1] == 1, t[:, 1], 1)

    results['t_flags'] = t_flags
    results['t_mle'] = t_mle

    # Create the objective function
    def fun(params, offset=False, lfp=False,
            zi=False, transform=True, gamma=0.):
        x_mle = np.copy(x)
        if transform:
            params = inv_trans(const(params))

        if offset:
            gamma = params[0]
            params = params[1:]
        else:
            # Use the assumed value
            pass

        if zi:
            f0 = params[-1]
            params = params[0:-1]
        else:
            f0 = 0.

        if lfp:
            p = params[-1]
            params = params[0:-1]
        else:
            p = 1.

        inf_c_flags, x_mle = _create_censor_flags(x_mle, gamma, c, dist)
        return dist.neg_ll(x_mle, c, n, inf_c_flags,
                           t_mle, t_flags, gamma, p, f0, *params)

    old_err_state = np.seterr(all='ignore')
    use_initial = False

    if zi:
        def jac(x, offset, lfp, zi, transform):
            return approx_fprime(x, fun, np.sqrt(np.finfo(float).eps),
                                 offset, lfp, zi, transform)
        hess = None
    else:
        jac = jacobian(fun)
        hess = hessian(fun)

    res = minimize(fun, init, args=(offset, lfp, zi, True),
                   method='Newton-CG', jac=jac, hess=hess)

    if (res.success is False) or (np.isnan(res.x).any()):
        res = minimize(fun, init, args=(offset, lfp, zi, True),
                       method='BFGS', jac=jac)

    if (res.success is False) or (np.isnan(res.x).any()):
        res = minimize(fun, init, args=(offset, lfp, zi, True))

    if 'Desired error ' in res['message']:
        res_tmp = minimize(fun, res.x, args=(offset, lfp, zi, True),
                           method='TNC', jac=jac)
        if res_tmp.success:
            res = res_tmp

    if 'Desired error not necessarily' in res['message']:
        print("Precision was lost, try:"
              + "\n- Using alternate fitting method"
              + "\n- visually checking model fit"
              + "\n- change data to be closer to 1.", file=sys.stderr)

    elif (not res.success) | (np.isnan(res.x).any()):
        print("MLE Failed: Try making the values of the data closer to "
              + "1 by dividing or multiplying by some constant."
              + "\n\nAlternately try setting the `init` keyword in the `fit()`"
              + " method to a value you believe is closer."
              + "A good way to do this is to set any shape parameter to 1. "
              + "and any scale parameter to be the mean of the data "
              + "(or it's inverse)"
              + "\n\nModel returned with inital guesses", file=sys.stderr)

        use_initial = True

    if use_initial:
        p_hat = inv_trans(const(init))
    else:
        p_hat = inv_trans(const(res.x))

    if offset:
        results['gamma'] = p_hat[0]
        params = p_hat[1:]
        parameters_for_hessian = copy.copy(params)
    else:
        results['gamma'] = 0
        params = p_hat
        parameters_for_hessian = copy.copy(params)

    if zi:
        results['f0'] = params[-1]
        params = params[0:-1]
    else:
        results['f0'] = 0.
        params = params

    if lfp:
        results['p'] = params[-1]
        results['params'] = params[0:-1]
    else:
        results['p'] = 1.
        results['params'] = params

    try:
        if zi or lfp:
            results['hess_inv'] = None
        else:
            results['hess_inv'] = inv(hess(parameters_for_hessian,
                                           *(False, lfp, zi,
                                             False, results['gamma'])))
    except np.linalg.LinAlgError:
        results['hess_inv'] = None

    results['_neg_ll'] = res['fun']
    results['res'] = res

    np.seterr(**old_err_state)

    return results



from surpyval import np
from scipy.stats import uniform
from scipy.special import ndtri as z

from surpyval import xcn_handler
from surpyval import nonparametric as nonp
from surpyval import parametric as para
from surpyval.parametric.parametric_fitter import ParametricFitter
from scipy.special import factorial
from .fitters.mpp import mpp

import warnings

class Exponential_(ParametricFitter):
    r"""

    Class used to generate the Exponential class.

    .. code:: python

        from surpyval import Exponential

    """
    def __init__(self, name):
        self.name = name
        self.k = 1
        self.bounds = ((0, None),)
        self.support = (0, np.inf)
        self.plot_x_scale = 'linear'
        self.y_ticks = [0.05, 0.4, 0.6, 0.7, 0.8, 0.9, 0.95, 0.99, 0.999, 0.9999]
        self.param_names = ['lambda']
        self.param_map = {
            'lambda' : 0,
        }

    def _parameter_initialiser(self, x, c=None, n=None, offset=False):
        # x, c, n = xcn_handler(x, c, n)
        # c = (c == 0).astype(np.int64)
        rate = 1./x[np.isfinite(x)].mean()
        if offset:
            return np.min(x) - (np.max(x) - np.min(x))/10., rate
        else:
            return np.array([rate])

    def sf(self, x, failure_rate):
        r"""

        Surival (or Reliability) function for the Exponential Distribution:

        .. math::
            R(x) = e^{-\lambda x}

        Parameters
        ----------

        x : numpy array or scalar
            The values at which the function will be calculated 
        failure_rate : numpy array or scalar
            The scale parameter for the Exponential distribution

        Returns
        -------

        sf : scalar or numpy array 
            The scalar value of the survival function of the distribution if a scalar was passed. If an array like object was passed then a numpy array is returned with the value of the survival function at each corresponding value in the input array.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Exponential
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> Exponential.sf(x, 3)
        array([4.97870684e-02, 2.47875218e-03, 1.23409804e-04, 6.14421235e-06,
               3.05902321e-07])
        """
        return np.exp(-failure_rate * x)

    def cs(self, x, X, failure_rate):
        r"""

        Conditional survival function for the Exponential Distribution:

        .. math::
            R(x) = e^{-\lambda x}

        The Exponential distribution is memoryless, and hence is the same as the regular survival distribution.

        Parameters
        ----------

        x : numpy array or scalar
            The value(s) at which the function will be calculated
        X : numpy array or scalar
            The value(s) at which each value(s) in x was known to have survived
        failure_rate : numpy array or scalar
            The scale parameter for the Exponential distribution

        Returns
        -------

        cs : scalar or numpy array 
            the conditional survival probability.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Exponential
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> Exponential.cs(x, 5, 3)
        array([4.97870684e-02, 2.47875218e-03, 1.23409804e-04, 6.14421235e-06,
               3.05902321e-07])
        """
        return self.sf(x, failure_rate)

    def ff(self, x, failure_rate):
        r"""

        CDF (or unreliability or failure) function for the Exponential Distribution:

        .. math::
            F(x) = 1 - e^{-\lambda x}

        Parameters
        ----------

        x : numpy array or scalar
            The values at which the function will be calculated 
        failure_rate : numpy array or scalar
            The scale parameter for the Exponential distribution

        Returns
        -------

        ff : scalar or numpy array 
            The value(s) of the CDF for each value of x.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Exponential
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> Exponential.ff(x, 3)
        array([0.95021293, 0.99752125, 0.99987659, 0.99999386, 0.99999969])
        """
        return 1 - np.exp(-failure_rate * x)

    def df(self, x, failure_rate):
        r"""

        Density function for the Exponential Distribution:

        .. math::
            f(x) = \lambda e^{-\lambda x}

        Parameters
        ----------

        x : numpy array or scalar
            The values at which the function will be calculated 
        failure_rate : numpy array or scalar
            The scale parameter for the Exponential distribution

        Returns
        -------

        df : scalar or numpy array 
            The density of the Exponential distribution for each value of x.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Exponential
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> Exponential.df(x, 3)
        array([1.49361205e-01, 7.43625653e-03, 3.70229412e-04, 1.84326371e-05,
               9.17706962e-07])
        """
        return failure_rate * np.exp(-failure_rate * x)

    def hf(self, x, failure_rate):
        r"""

        Instantaneous hazard rate for the Exponential Distribution.

        .. math::
            f(x) = \lambda

        The failure rate for the exponential distribution is constant. So this function only returns the input failure rate in the same shape as x.

        Parameters
        ----------

        x : numpy array or scalar
            The values at which the function will be calculated 
        failure_rate : numpy array or scalar
            The scale parameter for the Exponential distribution

        Returns
        -------

        hf : scalar or numpy array 
            The instantaneous hazard rate of the Exponential distribution for each value of x.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Exponential
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> Exponential.hf(x, 3)
        array([3, 3, 3, 3, 3])
        """
        return np.ones_like(x) * failure_rate

    def Hf(self, x, failure_rate):
        r"""

        Cumulative hazard rate for the Exponential Distribution.

        .. math::
            f(x) = \lambda x

        Parameters
        ----------

        x : numpy array or scalar
            The values at which the function will be calculated 
        failure_rate : numpy array or scalar
            The scale parameter for the Exponential distribution

        Returns
        -------

        hf : scalar or numpy array 
            The cumulative hazard rate of the Exponential distribution for each value of x.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Exponential
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> Exponential.Hf(x, 3)
        array([ 3,  6,  9, 12, 15])
        """
        return failure_rate * x

    def qf(self, p, failure_rate):
        r"""

        Quantile function for the Exponential Distribution:

        .. math::
            q(p) = \frac{-\ln\left ( p \right )}{\lambda}

        Parameters
        ----------

        p : numpy array or scalar
            The percentiles at which the quantile will be calculated 
        failure_rate : numpy array or scalar
            The scale parameter for the Exponential distribution

        Returns
        -------

        q : scalar or numpy array 
            The quantiles for the Exponential distribution at each value p.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Exponential
        >>> p = np.array([.1, .2, .3, .4, .5])
        >>> Exponential.qf(p, 3)
        array([0.76752836, 0.5364793 , 0.40132427, 0.30543024, 0.23104906])
        """
        return -np.log(p)/failure_rate

    def mean(self, failure_rate):
        r"""

        Calculates the mean of the Exponential distribution with given parameters.

        .. math::
            E = \frac{1}{\lambda}

        Parameters
        ----------

        failure_rate : numpy array or scalar
            The scale parameter for the Exponential distribution

        Returns
        -------

        mean : scalar or numpy array 
            The mean(s) of the Exponential distribution 

        Examples
        --------
        >>> from surpyval import Exponential
        >>> Exponential.mean(3)
        0.3333333333333333
        """
        return 1. / failure_rate

    def moment(self, n, failure_rate):
        r"""

        Calculates the n-th moment of the Exponential distribution.

        .. math::
            E = \frac{n!}{\lambda^{n}}

        Parameters
        ----------

        n : integer or numpy array of integers
            The ordinal of the moment to calculate
        failure_rate : numpy array or scalar
            The scale parameter for the Exponential distribution

        Returns
        -------

        moment : scalar or numpy array 
            The moment(s) of the Exponential distribution

        Examples
        --------
        >>> from surpyval import Exponential
        >>> Exponential.moment(2, 3)
        0.2222222222222222
        """
        return factorial(n) / (failure_rate ** n)

    def entropy(self, failure_rate):
        r"""

        Calculates the entropy of the Exponential distribution.

        .. math::
            S = 1 - \ln \left ( \lambda \right )

        Parameters
        ----------

        failure_rate : numpy array or scalar
            The scale parameter for the Exponential distribution

        Returns
        -------

        entropy : scalar or numpy array 
            The entropy(ies) of the Exponential distribution

        Examples
        --------
        >>> from surpyval import Exponential
        >>> Exponential.entropy(3)
        -0.09861228866810978
        """
        return 1 - np.log(failure_rate)

    def random(self, size, failure_rate):
        r"""

        Draws random samples from the distribution in shape `size`

        Parameters
        ----------

        size : integer or tuple of positive integers
            Shape or size of the random draw
        failure_rate : numpy array or scalar
            The scale parameter for the Exponential distribution

        Returns
        -------

        random : scalar or numpy array 
            Random values drawn from the distribution in shape `size`

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Exponential
        >>> Exponential.random(10, 3)
        array([0.32480264, 0.03186663, 0.41807108, 0.74221745, 0.06133774,
               0.2128422 , 0.36299424, 0.12250138, 0.61431089, 0.02266754])
        >>> Exponential.random((5, 5), 3)
        array([[0.25425552, 0.16867629, 0.21692401, 0.07020826, 0.03676643],
               [0.65528908, 0.20774767, 0.00625475, 0.04122388, 0.07089254],
               [1.22844679, 0.36199751, 0.564159  , 1.86811492, 0.08132478],
               [0.33541878, 0.38614518, 0.09285907, 0.33422975, 0.32515494],
               [0.03529228, 0.63134988, 0.45528738, 0.05037512, 0.7338039 ]])
        """
        U = uniform.rvs(size=size)
        return self.qf(U, failure_rate)

    def mpp_x_transform(self, x, gamma=0):
        return x - gamma

    def mpp_y_transform(self, y, *params):
        mask = ((y == 0) | (y == 1))
        out = np.zeros_like(y)
        out[~mask] = -np.log(1 - y[~mask])
        out[mask] = np.nan
        return out

    def mpp_inv_y_transform(self, y, *params):
        return 1 - np.exp(y)

    def mpp(self, x, c=None, n=None, heuristic="Nelson-Aalen", rr='y', on_d_is_0=False, offset=False):
        assert rr in ['x', 'y']
        x_pp, r, d, F = nonp.plotting_positions(x, c=c, n=n, heuristic=heuristic)

        if not on_d_is_0:
            x_pp = x_pp[d > 0]
            F    = F[d > 0]
        
        # Linearise
        y_pp = self.mpp_y_transform(F)

        mask = np.isfinite(y_pp)
        if mask.any():
            warnings.warn("Some Infinite values encountered in plotting points and have been ignored.", stacklevel=2)
            y_pp = y_pp[mask]
            x_pp = x_pp[mask]

        if offset:
            if   rr == 'y':
                params = np.polyfit(x_pp, y_pp, 1)
            elif rr == 'x':
                params = np.polyfit(y_pp, x_pp, 1)
            failure_rate = params[0]
            offset = -params[1] * (1./failure_rate)
            return tuple([offset, failure_rate])
        else:
            if   rr == 'y':
                x_pp = x_pp[:,np.newaxis]
                failure_rate = np.linalg.lstsq(x_pp, y_pp, rcond=None)[0]
            elif rr == 'x':
                y_pp = y_pp[:,np.newaxis]
                mttf = np.linalg.lstsq(y_pp, x_pp, rcond=None)[0]
                failure_rate = 1. / mttf
            return tuple([failure_rate[0]])

    def lambda_cb(self, x, failure_rate, cv_matrix, alpha_ci=0.05):
        return failure_rate * np.exp(np.array([-1, 1]).reshape(2, 1) * (z(alpha_ci/2) * 
                                    np.sqrt(cv_matrix.item()) / failure_rate))

    # def R_cb(self, x, failure_rate, cv_matrix, alpha_ci=0.05):
        # return np.exp(-self.lambda_cb(x, failure_rate, cv_matrix, alpha_ci=alpha_ci) * x).T

Exponential = Exponential_('Exponential')


import numpy as np
from surpyval import nonparametric as nonp
from surpyval.utils import xcnt_handler, xcnt_to_xrd, xrd_handler


class NonParametricFitter():
    def _create_non_p_model(self, x, r, d, estimator, data=None):
        out = nonp.NonParametric()
        if data is not None:
            out.data = data
        out.x = x
        out.r = r
        out.d = d
        out.R = nonp.FIT_FUNCS[estimator](r, d)
        out.model = self.how
        out.F = 1 - out.R
        with np.errstate(all='ignore'):
            out.H = -np.log(out.R)

        out.greenwood = self._compute_var(out.R, r, d)
        return out

    def _compute_var(self, R, r, d):
        with np.errstate(all='ignore'):
            var = d / (r * (r - d))
            var = np.where(np.isfinite(var), var, np.nan)
            greenwood = np.cumsum(var)
        return greenwood

    def fit(self, x=None, c=None, n=None, t=None,
            xl=None, xr=None, tl=None, tr=None,
            turnbull_estimator='Fleming-Harrington',
            set_lower_limit=None):
        r"""

        The central feature to SurPyval's capability. This function aimed to
        have an API to mimic the simplicity of the scipy API. That is, to use a
        simple :code:`fit()` call, with as many or as few parameters as are
        needed.

        Parameters
        ----------

        x : array like, optional
            Array of observations of the random variables. If x is
            :code:`None`, xl and xr must be provided.

        c : array like, optional
            Array of censoring flag. -1 is left censored, 0 is observed, 1 is
            right censored, and 2 is intervally censored. If not provided will
            assume all values are observed.

        n : array like, optional
            Array of counts for each x. If data is proivded as counts, then
            this can be provided. If :code:`None` will assume each
            observation is 1.

        t : 2D-array like, optional
            2D array like of the left and right values at which the
            respective observation was truncated. If not provided it assumes
            that no truncation occurs.

        tl : array like or scalar, optional
            Values of left truncation for observations. If it is a scalar
            value assumes each observation is left truncated at the value.
            If an array, it is the respective 'late entry' of the observation.

        tr : array like or scalar, optional
            Values of right truncation for observations. If it is a scalar
            value assumes each observation is right truncated at the value.
            If an array, it is the respective right truncation value for each
            observation.

        xl : array like, optional
            Array like of the left array for 2-dimensional input of x. This is
            useful for data that is all intervally censored. Must be used with
            the :code:`xr` input.

        xr : array like, optional
            Array like of the right array for 2-dimensional input of x. This is
            useful for data that is all intervally censored. Must be used with
            the :code:`xl` input.

        turnbull_estimator : ('Nelson-Aalen', 'Kaplan-Meier', or
        'Fleming-Harrington'), str, optional
            If using the Turnbull heuristic, you can elect to use either the
            KM, NA, or FH estimator with the Turnbull estimates of r, and d.
            Defaults to FH.

        Returns
        -------

        model : NonParametric
            A parametric model with the fitted parameters and methods for all
            functions of the distribution using the fitted parameters.

        Examples
        --------
        >>> from surpyval import NelsonAalen, Weibull, Turnbull
        >>> import numpy as np
        >>> x = Weibull.random(100, 10, 4)
        >>> model = NelsonAalen.fit(x)
        >>> print(model)
        Non-Parametric SurPyval Model
        =============================
        Model            : Nelson-Aalen
        >>> Turnbull.fit(x, turnbull_estimator='Kaplan-Meier')
        Non-Parametric SurPyval Model
        =============================
        Model            : Turnbull
        Estimator        : Kaplan-Meier
        """
        x, c, n, t = xcnt_handler(x=x, c=c, n=n, t=t,
                                  tl=tl, tr=tr,
                                  xl=xl, xr=xr)

        data = {}
        data['x'] = x
        data['c'] = c
        data['n'] = n
        data['t'] = t

        if self.how == 'Turnbull':
            data['estimator'] = turnbull_estimator
            out = nonp.NonParametric()
            t_obj = nonp.turnbull(x, c, n, t, turnbull_estimator)
            for k, v in t_obj.items():
                setattr(out, k, v)

            out.data = data
            return out

        else:
            x, r, d = xcnt_to_xrd(x, c, n, t)
            estimator = self.how

        if set_lower_limit is not None:
            x = np.hstack([[set_lower_limit], x])
            r = np.hstack([[r[0]], r])
            d = np.hstack([[0], d])

        return self._create_non_p_model(x, r, d, estimator=estimator,
                                        data=data)

    def from_xrd(self, x, r, d):
        r"""
        The central feature to SurPyval's capability. This function aimed to
        have an API to mimic the simplicity of the scipy API. That is, to use a
        simple :code:`fit()` call, with as many or as few parameters as are
        needed.

        Parameters
        ----------

        x : array like, optional
            Array of observations of the random variables. If x is
            :code:`None`, xl and xr must be provided.

        r : array like, optional
            Array of at risk items. For each value of x the r array is
            the number of at risk items immediately prior to the failures
            at x.

        d : array like, optional
            Array of counts of deaths/failures at each x. For each value of x
            the d array is the number of deaths at x (can be zero).

        Returns
        -------

        model : NonParametric
            A non-parametric model with the survival curved estimated
            using the selected method.

        Examples
        --------
        >>> from surpyval import NelsonAalen, Weibull, Turnbull
        >>> import numpy as np
        >>> x = [1, 2, 3, 4, 5, 6]
        >>> r = [10, 8, 6, 4, 3, 2]
        >>> d = [2, 1, 1, 1, 1, 1]
        >>> model = NelsonAalen.from_xrd(x)
        >>> print(model)
        Non-Parametric SurPyval Model
        =============================
        Model            : Nelson-Aalen
        """
        if self.how == 'Turnbull':
            raise ValueError("Can't use from_xrd with Turnbull estimator")

        x, r, d = xrd_handler(x, r, d)

        return self._create_non_p_model(x, r, d, estimator=self.how)



import numpy as np
import numpy_indexed as npi

def handle_xicn(x, i=None, c=None, n=None):
    x = np.array(x)

    if i is None:
        i = np.ones_like(x)
    else:
        i = np.array(i)

    if n is None:
        n = np.ones_like(x)
    else:
        n = np.array(n)

    if c is None:
        c = np.zeros_like(x)
    else:
        c = np.array(c)

    if npi.group_by(i).sum((c == 1).astype(int))[1].max() > 1:
        raise ValueError("An item is censored more than once.")

    # Check that if censored, it is the highest value
    for ii in set(i):
        ci = c[i == ii]
        xi = x[i == ii]
        if 1 in ci:
            if np.any(xi[ci == 1] <= xi[ci != 1]):
                raise ValueError("Item {} has censored value lower than an event value".format(ii))

    return x, i, c, n

def xicn_to_xrd(x, i, c, n):
    # This function assumes that it has been handled by the above
    # handler.
    gb_x = npi.group_by(x)
    r = npi.group_by(i).max(x)[1]
    r = np.array([1 if xi in r else 0 for xi in x])
    r = gb_x.sum(r)[1]
    r = r[::-1].cumsum()[::-1]
    x, d = gb_x.sum(n * (c == 0).astype(int))

    return x, r, d





__version__ = "0.10.2"

from autograd import numpy as np

from surpyval.utils import (
    xcn_sort,
    xcn_handler,
    xcn_to_xrd,
    xrd_to_xcn,
    xcnt_handler,
    xcnt_to_xrd,
    fsli_to_xcn,
    fsl_to_xcn,
    fs_to_xcn,
    fs_to_xrd,
    round_sig,
    xcn_to_fs
)

from surpyval.parametric import (
    Gumbel,
    Uniform,
    Exponential,
    Weibull,
    ExpoWeibull,
    Normal, Gauss,
    LogNormal, Galton,
    Logistic,
    LogLogistic,
    Gamma,
    Beta,
    Distribution,
    MixtureModel
)

from surpyval.nonparametric import (
    KaplanMeier,
    NelsonAalen,
    FlemingHarrington,
    Turnbull
)

from surpyval.regression import (
    CoxPH
)

NUM = np.float64
TINIEST = np.finfo(np.float64).tiny
EPS = np.sqrt(np.finfo(NUM).eps)



from surpyval import np
from scipy.optimize import minimize


def mom_fun(params, dist, inv_trans, const, offset, moments):
    dist_moments = dist.mom_moment_gen(*inv_trans(const(params)),
                                       offset=offset)
    return np.log(((np.abs(moments - dist_moments) / moments))).sum()


def mom(model):
    """
    MOM: Method of Moments.

    This is one of the simplest ways to calculate the parameters of a
    distribution. This method is quick but only works with uncensored data.
    """
    dist = model.dist
    x, n = model.data['x'], model.data['n']

    const = model.fitting_info['const']
    inv_trans = model.fitting_info['inv_trans']
    init = model.fitting_info['init']
    offset = model.offset

    x_ = np.repeat(x, n)

    if hasattr(dist, '_mom'):
        return {'params': dist._mom(x_)}

    moments = np.zeros(model.k)

    for i in range(0, model.k):
        moments[i] = (x_**(i + 1)).mean()

    res = minimize(mom_fun, np.array(init), tol=1e-15,
                   args=(dist, inv_trans, const, offset, moments))

    params = inv_trans(const(res.x))

    results = {}
    if offset:
        results['gamma'] = params[0]
        results['params'] = params[1:]
    else:
        results['params'] = params

    results['res'] = res
    results['offset'] = offset

    return results



from autograd import jacobian, hessian
import autograd.numpy as np
import inspect
from autograd import elementwise_grad

from scipy.optimize import minimize
import surpyval

from ..parametric.fitters import bounds_convert, fix_idx_and_function
from .regression import Regression

from ..utils import _get_idx

class ProportionalHazardsModel():
    """
    A Proportional Hazard Model class.
    Currently only implemented for semi-parametric
    would need to change should we want to allow
    fully parametric models.
    # """
    def __init__(self, kind, parameterization):
        self.kind = kind
        self.parameterization = parameterization

    def __repr__(self):
        out = ('Regression SurPyval Model'
            + '\n========================='
            + '\nType                : Proportional Hazards'
            + '\nKind                : {kind}'
            + '\nParameterization    : {parameterization}'
            ).format(kind=self.kind,
                    parameterization=self.parameterization)
            
        out = (out + '\nParameters          :\n')
        for i, p in enumerate(self.parameters):
            out += '   beta_{i}  :  {p}\n'.format(i=i, p=p)
        return out

    def hf(self, x, Z):
        idx, rev = _get_idx(self.x, x)
        return (self.h0[idx] * self.phi(Z))[rev]

    def Hf(self, x, Z):
        idx, rev = _get_idx(self.x, x)
        return (self.H0[idx] * self.phi(Z))[rev]

    def sf(self, x, Z):
        return np.exp(-self.Hf(x, Z))

    def ff(self, x, Z):
        return 1 - self.sf(x, Z)

    def df(self, x, Z):
        return self.hf(x, Z) * self.sf(x, Z)


class ProportionalHazardsFitter():
    def __init__(self, name, dist, phi, phi_bounds, phi_param_map,
                 baseline=[], fixed={}, phi_init=None):

        if str(inspect.signature(phi)) != '(X, *params)':
            raise ValueError('PH function must have the signature \'(X, *params)\'')

        if type(baseline) != list:
            # If passed a single string..
            baseline = [baseline]

        self.name = name
        self.dist = dist
        self.k_dist = len(self.dist.param_names)
        self.bounds = self.dist.bounds
        self.support = self.dist.support
        self.param_names = self.dist.param_names
        self.param_map = {v : i for i, v in enumerate(self.dist.param_names)}
        self.phi = phi
        self.Hf_dist = self.dist.Hf
        self.hf_dist = self.dist.hf
        self.sf_dist = self.dist.sf
        self.ff_dist = self.dist.ff
        self.df_dist = self.dist.df
        self.baseline = baseline
        self.fixed = fixed
        self.phi_init = phi_init
        self.phi_bounds = phi_bounds
        self.phi_param_map = phi_param_map

    def Hf(self, x, X, *params):
        dist_params = np.array(params[0:self.k_dist])
        phi_params = np.array(params[self.k_dist:])
        Hf_raw = self.Hf_dist(x, *dist_params)
        return self.phi(X, *phi_params) * Hf_raw

    def hf(self, x, X, *params):
        dist_params = np.array(params[0:self.k_dist])
        phi_params = np.array(params[self.k_dist:])
        hf_raw = self.hf_dist(x, *dist_params)
        return self.phi(X, *phi_params) * hf_raw

    def df(self, x, X, *params):
        return self.hf(x, X, *params) * np.exp(-self.Hf(x, X, *params))

    def sf(self, x, X, *params):
        return np.exp(-self.Hf(x, X, *params))

    def ff(self, x, X, *params):
        return 1 - np.exp(-self.Hf(x, X, *params))

    def _parameter_initialiser_dist(self, x, c=None, n=None, t=None):
        out = []
        for low, high in self.bounds:
            if (low is None) & (high is None):
                out.append(0)
            elif high is None:
                out.append(low + 1.)
            elif low is None:
                out.append(high - 1.)
            else:
                out.append((high + low)/2.)

        return out

    def mpp_inv_y_transform(self, y, *params):
        return y

    def mpp_y_transform(self, y, *params):
        return y

    def mpp_inv_y_transform(self, y, *params):
        return y

    def mpp_x_transform(self, x, gamma=0):
        return x - gamma

    def log_df(self, x, X, *params):
        return np.log(self.hf(x, X, *params)) - self.Hf(x, X, *params)

    def log_sf(self, x, X, *params):
        return -self.Hf(x, X, *params)

    def log_ff(self, x, X, *params):
        return np.log(self.ff(x, X, *params))

    def random(self, size, X, *params):
        dist_params = np.array(params[0:self.k_dist])
        phi_params = np.array(params[self.k_dist:])
        random = []
        U = np.random.uniform(0, 1, size)
        x = self.dist.qf(U**(self.phi(X, *phi_params)), *dist_params)
        X_out = np.ones_like(x) * X
        return x.flatten(), X_out.flatten()

    def neg_ll(self, X, x, c, n, *params):
        params = np.array(params)
        like = np.zeros_like(x).astype(float)
        like = np.where(c ==  0, self.log_df(x, X, *params), like)
        like = np.where(c ==  1, self.log_sf(x, X, *params), like)
        like = np.where(c ==  -1, self.log_ff(x, X, *params), like)
        like = np.multiply(n, like)
        return -np.sum(like)

    def fit(self, X, x, c=None, n=None, t=None, init=[], fixed={}):
        x, c, n, t = surpyval.xcnt_handler(x, c, n, t, group_and_sort=False)

        if init == []:
            ps = self.dist.fit(x, c=c, n=n, t=t).params
            if callable(self.phi_init):
                init_phi = self.phi_init(X)

            init = np.array([*ps, *init_phi])
        else:
            init = np.array(init)

        if self.baseline != []:
            baseline_model = self.dist.fit(x, c, n, t)
            baseline_fixed = {k : baseline_model.params[baseline_model.param_map[k]] for k in self.baseline}
            fixed = {**baseline_fixed, **fixed}

        # Dynamic or static bounds determination
        if callable(self.phi_bounds):
            bounds = (*self.bounds, *self.phi_bounds(X))
        else:
            bounds = (*self.bounds, *self.phi_bounds)

        if callable(self.phi_param_map):
            phi_param_map = self.phi_param_map(X)
        else:
            phi_param_map = self.phi_param_map

        param_map = {**self.param_map, **phi_param_map}

        transform, inv_trans, funcs, inv_f = bounds_convert(x, bounds)
        const, fixed_idx, not_fixed = fix_idx_and_function(fixed, param_map, funcs)

        init = transform(init)[not_fixed]

        with np.errstate(all='ignore'):
            fun  = lambda params : self.neg_ll(X, x, c, n, *inv_trans(const(params)))
            # jac = jacobian(fun)
            # hess = hessian(fun)
            res = minimize(fun, init)
            res = minimize(fun, res.x, method='TNC')

        params = inv_trans(const(res.x))

        model = Regression()
        model.model = self
        model.reg_model = self.phi
        model.kind = "Proportional Hazard"
        model.distribution = self.dist
        model.params = np.array(params)
        model.res = res
        model._neg_ll = res['fun']
        model.fixed = self.fixed
        model.k_dist = self.k_dist
        model.phi_param_map = phi_param_map

        print(res)

        return model



"""
This code was created for and sponsored by Cartiga (www.cartiga.com). 
Cartiga makes no representations or warranties in connection with the code 
and waives any and all liability in connection therewith. Your use of the 
code constitutes acceptance of these terms.

Copyright 2022 Cartiga LLC
"""

import numpy as np
from .fine_gray import FineGray
from surpyval import CoxPH
from surpyval.utils import _get_idx
from surpyval.utils import (
    validate_fine_gray_inputs, 
    _scale
)
from scipy.optimize import minimize

class CompetingRiskProportionalHazard():
    """
    TODO: Time-Varying Implementation
    TODO: Change this to SemiParametricCompetingRiskProportionalHazard ??
    """

    def _f(self, arr, x, Z, event=None, interp='step'):
        idx, rev = _get_idx(self.x, x)

        if event is not None and event not in self.event_idx_map:
            raise ValueError("Unrecognised event type for this model")

        if event is None:
            return (arr.sum(axis=0)[idx] * self.phi(Z))[rev]
        else:
            e_i = self.event_idx_map.get(event, None)
            return (arr[e_i, idx] * self.phi_e(Z, e_i))[rev]

    def hf(self, x, Z, event=None, interp='step'):
        return self._f(self.h0_e, x, Z, event=event, interp=interp)

    def Hf(self, x, Z, event=None, interp='step'):
        return self._f(self.H0_e, x, Z, event=event, interp=interp)

    def sf(self, x, Z, event=None, interp='step'):
        return np.exp(-self.Hf(x, Z, event=event, interp=interp))

    def ff(self, x, Z, event=None, interp='step'):
        return 1 - self.sf(x, Z, event=event, interp=interp)

    def df(self, x, Z, event=None, interp='step'):
        return (self.hf(x, Z, event=event, interp=interp)
                * self.sf(x, Z, event=event, interp=interp))

    def cif(self, x, Z, event):
        # Index and reverse index
        # in case x is not in order.
        idx, rev = _get_idx(self.x, x)

        # CIF
        lambda_e = self.hf(self.x, Z, event)
        S = self.sf(self.x, Z)
        # iif = instantaneous incidence function
        iif = lambda_e * S
        cif = iif.cumsum()
        
        return cif[idx][rev]

    
    @classmethod
    def cox_risk_set_indices(cls, x_i, e_i, x, e):
        return (x >= x_i)

    @classmethod
    def fine_gray_risk_set_indices(cls, x_i, e_i, x, e):
        return ((x >= x_i) | (e != e_i))

    @classmethod
    def partial_log_like(cls, beta, x, c, n, Z, e, 
                         event, how='Cox', scale=False):
        """
        This is the Breslow implementation
        TODO: 
        - Efron, and
        - Kalbfleisch and Prentice (This is what we need!)
        """
        # print(beta)
        if how == 'Cox':
            risk_set_indices = cls.cox_risk_set_indices
        elif how == 'Fine-Gray':
            risk_set_indices = cls.fine_gray_risk_set_indices
        N = len(x)

        ll = np.zeros_like(x)
        idx = np.array(range(N))

        for i, x_i in enumerate(x):
            if c[i] == 1:
                continue
            elif e[i] != event:
                continue
            # This is the key insight: 
            # all non e failures are still at risk!
            at_risk = risk_set_indices(x_i, event, x, e)
            Z_ri = Z[at_risk, :]
            # Sum of 'Z's is sometimes also called 'S'
            # e.g. see https://myweb.uiowa.edu/pbreheny/7210/f15/notes/11-5.pdf
            Z_i = n[i] * Z[i, :]
            # Breslow log-like
            ll_i = (Z_i @ beta) - (n[i] * np.log(np.exp(Z_ri @ beta).sum()))
            ll = np.where(i == idx, ll_i, ll)

        return _scale(ll, n, scale)

    @classmethod
    def baseline(cls, beta, x, c, n, Z, e, event):
        unique_x = np.unique(x)

        d = np.zeros_like(unique_x)
        r = np.zeros_like(unique_x)
        for i, tau_i in enumerate(unique_x):
            mask_d_i = (x == tau_i) & (c == 0)
            d[i] = n[mask_d_i].sum()

            mask_at_risk_i = (x >= tau_i) | (e != event)
            Z_ri = Z[mask_at_risk_i, :]

            r[i] = np.exp(Z_ri @ beta).sum()

        return d / r

    @classmethod
    def fit_from_df(self, *args, **kwargs):
        # TODO: Finish this
        raise NotImplementedError("Not yet...")

    @classmethod
    def fit(cls, x, Z, e, c=None, n=None, how="Cox", tie_method="efron"):
        r"""
        This function aimed to have an API to mimic the simplicity 
        of the scipy API. That is, to use a simple :code:`fit()` call, 
        with as many or as few parameters as are needed.

        API is plaigiarised from surpyval (which I also authored :) )

        Parameters
        ----------

        x : array like
            Array of observations of the random variables.

        Z : ndarray like
            Array of covariates for each random variable, x.
        
        e : array like
            Array of events that corresponds to each x.

        c : array like, optional
            Array of censoring flag. -1 is left censored, 0 is observed, 1 is
            right censored, and 2 is intervally censored. Only right censored
            data is implemented in FineGray. In not provided assumes each x
            was fully observed, i.e. c is automatically set to 0 for all x.

        n : array like, optional
            Array of counts for each x. If data is proivded as counts, then
            this can be provided. If :code:`None` will assume each
            observation is 1.

        method : str, optional
            String which declares method which is used to estimate the 
            baseline survival function. Can be either 'Nelson-Aalen' or
            'Kaplan-Meier'. Default is 'Nelson-Aalen'.

        Returns
        -------

        model : CompetingRiskProportionalHazard
            A Competing Risk Proportional Hazard model with fitted params
            and helper methods using the fitted params.

        Examples
        --------
        >>> from cartiga import CRPH
        
        """
        x, Z, e, c, n = validate_fine_gray_inputs(x, Z, e, c, n)

        unique_e = set(e)
        if None in unique_e:
            unique_e.remove(None)

        n_event_types = len(unique_e)

        event_idx_map = {state : i for i, state in enumerate(unique_e)}

        betas = np.zeros((len(unique_e), Z.shape[1]))
        unique_x = np.unique(x)

        baselines = np.zeros((len(unique_e), len(unique_x)))
        # Best initial assumption is to assume there is no risk
        # beta_init = np.zeros(Z.shape[1])

        model = cls()
        model.n_event_types = n_event_types
        model.event_idx_map = event_idx_map
        model.how = how

        if how == 'Cox':
            # The Cox method is just assuming the other methods
            # are right censored.
            results = []
            for i, event in enumerate(unique_e):
                c_e = np.where(e == event, 0, 1)
            
                res = CoxPH.fit(x, Z, c_e, n,
                                method=tie_method,
                                with_hess=False).res
                                
                results.append(res)
                betas[i, :] = res.x
                baselines[i, :] = cls.baseline(res.x, x, c, n, Z, e, event)

        elif how == 'Fine-Gray':
            results, unique_e = FineGray.fit(x, Z, e, c, n)
            for i, res in enumerate(results):
                betas[i, :] = res.x
                baselines[i, :] = cls.baseline(res.x, x, c, n, Z, e, unique_e[i])
        else:
            raise ValueError("`how` must be either 'Cox' or 'Fine-Gray")

        model.results = results
        model.betas = betas
        model.beta = betas.sum(axis=0)
        model.phi_e = lambda Z, e_i: np.exp(Z @ model.betas[e_i, :])
        model.phi = lambda Z: np.exp(Z @ model.beta)
        model.h0_e = baselines
        model.H0_e = baselines.cumsum(axis=1)
        model.x = unique_x
        return model

CRPH = CompetingRiskProportionalHazard


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="surpyval",
    version="0.10.6",
    author="Derryn Knife",
    author_email="derryn@reliafy.com",
    description="A python package for survival analysis",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/derrynknife/SurPyval",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=['autograd', 'numpy', 'scipy', 'pandas','autograd_gamma',
                      'matplotlib', 'numpy_indexed', 'numba', 'formulaic'],
    include_package_data=True,
    package_data={'': ['datasets/*.csv']},
)


import numpy as np
from surpyval.nonparametric.nonparametric_fitter import NonParametricFitter


def nelson_aalen(r, d):
    H = np.cumsum(d / r)
    H[np.isnan(H)] = np.inf
    R = np.exp(-H)
    return R


class NelsonAalen_(NonParametricFitter):
    r"""
    Nelson-Aalen estimator class. Returns a `NonParametric`
    object from method :code:`fit()` Calculates the Non-Parametric
    estimate of the survival function using:

    .. math::
        R(x) = e^{-\sum_{i:x_{i} \leq x}^{} \frac{d_{i} }{r_{i}}}

    Examples
    --------
    >>> import numpy as np
    >>> from surpyval import NelsonAalen
    >>> x = np.array([1, 2, 3, 4, 5])
    >>> model = NelsonAalen.fit(x)
    >>> model.R
    array([0.81873075, 0.63762815, 0.45688054, 0.27711205, 0.10194383])
    """
    def __init__(self):
        self.how = 'Nelson-Aalen'


NelsonAalen = NelsonAalen_()



import pandas as pd
from pkg_resources import resource_filename


class BoforsSteel_():
    r"""
    A Class with a Pandas DataFrame containing the data of
    the tensile strenght of Bofors Steel from the Weibull paper [1]_.

    Attributes
    ----------
    df : DataFrame
        A Pandas DataFrame containing the data of
        the tensile strenght of Bofors Steel from the Weibull paper.

    Examples
    --------
    >>> from surpyval import BoforsSteel
    >>> df = BoforsSteel.df
    >>> df.head(5)

    .. list-table::
        :header-rows: 1

        * -
          - x
          - n
        * - 0
          - 40.800
          - 10
        * - 1
          - 42.075
          - 23
        * - 2
          - 43.350
          - 48
        * - 3
          - 44.625
          - 80
        * - 4
          - 45.900
          - 63

    References
    ----------
    .. [1] Weibull, W., A statistical distribution function
           of wide applicability, Journal of applied mechanics,
           Vol. 18, No. 3, pp 293-297 (1951).

    """
    def __init__(self):
        self.data = pd.read_csv(resource_filename("surpyval",
                                                "datasets/bofors_steel.csv"),
                              engine="python")
    
    def __repr__(self):
        return """
        Data from:
        Weibull, W., A Statistical Distribution Function
        of Wide Applicability, Journal of Applied Mechanics,
        Vol. 18, No. 3, pp 293-297 (1951).
        https://doi.org/10.1115/1.4010337
        """


class Boston_():
    def __init__(self):
        self.data = pd.read_csv(resource_filename("surpyval",
                                                "datasets/boston.csv"),
                              engine="python")


class Bearing_():
    def __init__(self):
        x = [17.88, 28.92, 33, 41.52, 42.12, 45.6, 48.4, 51.84,
             51.96, 54.12, 55.56, 67.8, 68.64, 68.64, 68.88, 84.12,
             93.12, 98.64, 105.12, 105.84, 127.92, 128.04, 173.4]
        self.data = pd.DataFrame({'Cycles to Failure (millions)': x})

    def __repr__(self):
        return """
        Data from:
        Lieblein, J. and Zelen, M. (1956) Statistical Investigation
        of the Fatigue Life of Deep-Groove Ball Bearings. Journal of
        Research of the National Bureau of Standards, 57, 273-315.
        http://dx.doi.org/10.6028/jres.057.033
        """

class Heart_():
  def __init__(self):
    self.data = pd.read_csv(resource_filename("surpyval",
                                                "datasets/heart.csv"),
                              engine="python")
  
  def __repr__(self):
    return """
    Data from:
    Heart transplant data
    """

class Lung_():
  def __init__(self):
    self.data = pd.read_csv(resource_filename("surpyval",
                                                "datasets/lung.csv"),
                              engine="python")
  
  def __repr__(self):
    return """
    Data from:
    Lung data...
    """

class Rossi_():
  def __init__(self):
    self.data = pd.read_csv(resource_filename("surpyval",
                                                "datasets/rossi.csv"),
                              engine="python")
    self.time_varying_data = pd.read_csv(resource_filename("surpyval",
                                                "datasets/rossi_tv.csv"),
                              engine="python")
  
  def __repr__(self):
    return """
    Data from:
    Rossi
    """

class Tires_():
  def __init__(self):
    self.data = pd.read_csv(resource_filename("surpyval",
                                              "datasets/tires.csv"),
                            engine="python")
  
  def __repr__(self):
    return """
    Data from:
    V.V Krivtsov, D.E Tananko, T.P Davis,
    Regression Approach to Tire Reliability Analysis,
    Reliability Engineering & System Safety,
    Volume 78, Issue 3, 2002,
    Pages 267-273,
    https://doi.org/10.1016/S0951-8320(02)00169-2.
    """


BoforsSteel = BoforsSteel_()
Bearing = Bearing_()
Boston = Boston_()
Heart = Heart_()
Lung = Lung_()
Rossi = Rossi_()
Tires = Tires_()



import numpy as np
from surpyval.nonparametric.nonparametric_fitter import NonParametricFitter


def fh_h(r_i, d_i):
    out = 0
    while d_i > 1:
        out += 1. / r_i
        r_i -= 1
        d_i -= 1
    out += d_i / r_i
    return out


def fleming_harrington(r, d):
    Y = np.array([fh_h(r_i, d_i) for r_i, d_i in zip(r, d)])
    H = Y.cumsum()
    H[np.isnan(H)] = np.inf
    R = np.exp(-H)
    return R


class FlemingHarrington_(NonParametricFitter):
    r"""
    Fleming-Harrington estimation of survival distribution.
    Returns a `NonParametric` object from method :code:`fit()`
    calculates the Non-Parametric estimate of the survival function using:

    .. math::

        R = e^{-\sum_{i:x_{i} \leq x} \sum_{i=0}^{d_x-1} \frac{1}{r_x - i}}

    See 'NonParametric section for detailed estimate of how H is computed.'

    Examples
    --------
    >>> import numpy as np
    >>> from surpyval import FlemingHarrington
    >>> x = np.array([1, 2, 3, 4, 5])
    >>> model = FlemingHarrington.fit(x)
    >>> model.R
    array([0.81873075, 0.63762815, 0.45688054, 0.27711205, 0.10194383])
    """
    def __init__(self):
        self.how = 'Fleming-Harrington'


FlemingHarrington = FlemingHarrington_()



import numpy as np


def success_run(n, confidence=0.95, alpha=None):
    r"""
    Function that can be used to estimte the confidence given n samples
    all survive a test.
    """
    if alpha is None:
        alpha = 1 - confidence
    return np.power(alpha, 1. / n)



"""

Parametric Analysis
===================

.. code:: python

	import surpyval.parametric as para

	model = para.Weibull.fit(x)

"""
from surpyval import nonparametric as nonp
from .parametric_fitter import ParametricFitter
from .parametric import Parametric
from .gumbel import Gumbel
from .weibull import Weibull
from .exponential import Exponential
from .expo_weibull import ExpoWeibull
from .normal import Normal
from .normal import Gauss
from .lognormal import LogNormal
from .lognormal import Galton
from .gamma import Gamma
from .beta import Beta
from .uniform import Uniform
from .logistic import Logistic
from .loglogistic import LogLogistic
from .mixture_model import MixtureModel
from .dist_from_hazard import Distribution




import autograd.numpy as np
from collections import defaultdict

from .cox_ph import CoxPH

from .proportional_hazards import ProportionalHazardsFitter
from .accelerated_failure_time import AcceleratedFailureTimeFitter
from .parameter_substitution import ParameterSubstitutionFitter

import surpyval as surv
from .lifemodels import (
    InversePower, Power,
    InverseExponential, Exponential,
    Linear,
    Eyring, InverseEyring,
    DualExponential,
    DualPower,
    PowerExponential,
    GeneralLogLinear
)

from ..parametric import (
    LogNormal,
    Normal,
    Weibull,
    Gumbel,
    Logistic
)

# Useful for proportional odds
# https://data.princeton.edu/pop509/parametricsurvival.pdf

# Semi-Parametric Proportional Hazard
# CoxPH = CoxProportionalHazardsFitter()

DISTS = [surv.Exponential, Normal, Weibull, Gumbel, Logistic, LogNormal]
LIFE_PARAMS = ['lambda', 'mu', 'alpha', 'mu', 'mu', 'mu']
LIFE_MODELS = [
    Power, 
    InversePower,
    Exponential,
    InverseExponential,
    Eyring,
    InverseEyring,
    Linear,
    DualExponential,
    DualPower,
    PowerExponential
]

life_parameter_transform = defaultdict(lambda: None)
life_parameter_inverse_transform = defaultdict(lambda: None)
baseline_parameters = defaultdict(lambda: [])

life_parameter_transform['LogNormal'] = lambda x: np.log(x)
life_parameter_transform['Exponential'] = lambda x: 1./x

life_parameter_inverse_transform['LogNormal'] = lambda x: np.exp(x)
life_parameter_inverse_transform['Exponential'] = lambda x: 1./x

# Quite remarkable - creates every life model and distribution class!
for dist, parameter in zip(DISTS, LIFE_PARAMS):
    for life_model in LIFE_MODELS:
        name = dist.name + life_model.name + "AL"
        vars()[name] = ParameterSubstitutionFitter(
                            'Accelerated Life', 
                            name, dist, life_model, 
                            parameter,
                            param_transform=life_parameter_transform[dist.name],
                            inverse_param_transform=life_parameter_inverse_transform[dist.name]
                        )

for dist, parameter in zip(DISTS, LIFE_PARAMS):
    name = dist.name + life_model.name + "AL"
    vars()[name] = ParameterSubstitutionFitter(
                        'Accelerated Life', 
                        name,
                        dist,
                        GeneralLogLinear,
                        parameter,
                        baseline=[parameter],
                        param_transform=life_parameter_transform[dist.name],
                        inverse_param_transform=life_parameter_inverse_transform[dist.name]
                    )

# Parametric Proportional Hazard
ExponentialPH = ProportionalHazardsFitter(
    'ExponentialPH', 
    surv.Exponential,
    lambda X, *params: np.exp(np.dot(X, np.array(params))),
    lambda X: (((None, None),) * X.shape[1]),
    phi_param_map=lambda X: {'beta_' + str(i) : i for i in range(X.shape[1])},
    baseline=['lambda'],
    phi_init=lambda X: np.zeros(X.shape[1])
)


# Parametric AFT
WeibullInversePowerAFT = AcceleratedFailureTimeFitter(
    'WeibullInversePowerAFT', 
    surv.Weibull,
    InversePower
)
